"""This is the unit test driver module."""
import sys
from pathlib import Path

# Ensure the local source tree takes precedence over any installed version
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))

# Force reimport of batcave from the local source
if 'batcave' in sys.modules:
    # Remove all batcave submodules from the cache
    for mod_name in list(sys.modules):
        if mod_name == 'batcave' or mod_name.startswith('batcave.'):
            del sys.modules[mod_name]



"""This module provides utilities for working with servers.

Attributes:
    _STATUS_CHECK_INTERVAL (int, default=2): This is the default wait in seconds when performing any checks.
"""
# pylint: disable=invalid-name,too-many-lines

# Import standard modules
import sys
from csv import DictReader
from enum import Enum
from os import environ
from pathlib import Path
from platform import node
from socket import getfqdn, gethostbyname, gaierror
from string import Template
from time import sleep
from typing import cast, List, Optional, Tuple, Union
from xml.etree.ElementTree import Element, SubElement, parse as xml_parse

# Import third-party modules
from psutil import process_iter, NoSuchProcess, Process as _LinuxProcess

# Import internal modules
from .platarch import OsType
from .serverpath import ServerPath  # pylint: disable=cyclic-import
from .sysutil import syscmd, CMDError, ServerAuthType
from .lang import BatCaveError, BatCaveException, CommandResult, PathName, WIN32

if sys.platform == 'win32':
    from pywintypes import com_error  # pylint: disable=no-name-in-module,import-error
    from win32com.client import CDispatch, DispatchEx  # pylint: disable=import-error
    from wmi import connect as WMI, x_wmi  # pylint: disable=import-error
    from .iispy import IISInstance
    _DEFAULT_WMI = True
else:
    _DEFAULT_WMI = False

    class x_wmi(Exception):
        'Needed to avoid errors on Linux'

    class WMI:  # pylint: disable=too-few-public-methods
        'Needed to avoid errors on Linux'
        def __init__(self, *args, **kwargs):
            pass

_STATUS_CHECK_INTERVAL = 2

ProcessSignal = Enum('ProcessSignal', ('stop', 'kill'))
ServiceSignal = Enum('ServiceSignal', ('disable', 'enable', 'start', 'stop', 'pause', 'resume', 'restart'))
ServiceState = Enum('ServiceState', ('StartPending', 'ContinuePending', 'Running', 'StopPending', 'Stopped', 'PausePending', 'Paused'))
ServiceType = Enum('ServiceType', ('systemd', 'sysv', 'upstart', 'windows'))
TaskSignal = Enum('TaskSignal', ('enable', 'disable', 'run', 'end'))

type ServerType = Union[str, 'Server']
type ServerManager = Union[WMI, 'OSManager']
type WMIObject = Union[bool, WMI]


class ServerObjectManagementError(BatCaveException):
    """Server Exceptions.

    Attributes:
        BAD_FILTER: Multiple filters were specified when only one is allowed.
        BAD_OBJECT_SIGNAL: The signal type is invalid.
        BAD_OBJECT_STATE: The object state is unknown.
        BAD_TRANSITION: The object state transition was unexpected.
        OBJECT_NOT_FOUND: The requested object was not found.
        REMOTE_CONNECTION_ERROR: The was an error attempting to connect to the remote server.
        REMOTE_NOT_SUPPORTED: The request action is not supported for remote servers.
        SERVER_NOT_FOUND: The requested server was not found.
        NOT_UNIQUE: A unique instance of the object was not found.
        STATUS_CHECK_TIMEOUT: The was a timeout checking for the status of the object.
        WMI_ERROR: The was a WMI error.
    """
    BAD_FILTER = BatCaveError(1, 'One and only one filter must be provided for this object')
    BAD_OBJECT_SIGNAL = BatCaveError(2, Template('Unknown object signal: $signal'))
    BAD_OBJECT_STATE = BatCaveError(3, Template('Unknown object state: $state'))
    BAD_TRANSITION = BatCaveError(4, Template('Invalid state transition from $from_state to $to_state'))
    NOT_UNIQUE = BatCaveError(5, Template('Unable to locate unique $type according to $filters'))
    OBJECT_NOT_FOUND = BatCaveError(6, Template('No $type object found'))
    REMOTE_CONNECTION_ERROR = BatCaveError(7, Template('Unable to connect to remote server $server: $msg'))
    REMOTE_NOT_SUPPORTED = BatCaveError(8, 'Remote objects are not supported for Linux servers')
    SERVER_NOT_FOUND = BatCaveError(9, Template('No server found: $server'))
    STATUS_CHECK_TIMEOUT = BatCaveError(10, Template('Timeout waiting for expected $type state: $state'))
    WMI_ERROR = BatCaveError(11, Template('WMI error on server $server: $msg'))


class Server:
    """Class to create a universal abstract interface for a server.

    Attributes:
        _WSA_NAME_OR_SERVICE_NOT_KNOWN: Error code indicating the service is unknown.
        _WSAHOST_NOT_FOUND: Error code indicating the host was not found.
        _WMI_SERVICE_CREATE_ERRORS: A dictionary to map WMI errors to errors messages.
    """
    _WMI_SERVICE_CREATE_ERRORS = {1: 'The request is not supported.',
                                  2: 'The user did not have the necessary access.',
                                  3: 'The service cannot be stopped because other services that are running are dependent on it.',
                                  4: 'The requested control code is not valid, or it is unacceptable to the service.',
                                  5: 'The requested control code cannot be sent to the service because the state of the service (State property of the Win32_BaseService class) is equal to 0, 1, or 2.',  # noqa: E501, pylint: disable=line-too-long
                                  6: 'The service has not been started.',
                                  7: 'The service did not respond to the start request in a timely fashion.',
                                  8: 'Unknown failure when starting the service.',
                                  9: 'The directory path to the service executable file was not found.',
                                  10: 'The service is already running.',
                                  11: 'The database to add a new service is locked.',
                                  12: 'A dependency this service relies on has been removed from the system.',
                                  13: 'The service failed to find the service needed from a dependent service.',
                                  14: 'The service has been disabled from the system.',
                                  15: 'The service does not have the correct authentication to run on the system.',
                                  16: 'This service is being removed from the system.',
                                  17: 'The service has no execution thread.',
                                  18: 'The service has circular dependencies when it starts.',
                                  19: 'A service is running under the same name.',
                                  20: 'The service name has invalid characters.',
                                  21: 'Invalid parameters have been passed to the service.',
                                  22: 'The account under which this service runs is either invalid or lacks the permissions to run the service.',
                                  23: 'The service exists in the database of services available from the system.',
                                  24: 'The service is currently paused in the system.'}
    _WSA_NAME_OR_SERVICE_NOT_KNOWN = -2
    _WSAHOST_NOT_FOUND = 11001

    def __init__(self, hostname: Optional[str] = None, domain: Optional[str] = None, *, auth: Optional[ServerAuthType] = None, defer_wmi: bool = True, ip: str = '',
                 os_type: OsType = OsType.windows if WIN32 else OsType.linux):
        """
        Args:
            hostname (optional): The server hostname. If not specified, will default to the name of the localhost.
            domain (optional): The server domain. If not specified, will default to the domain of the localhost.
            auth (optional, default=None): A tuple of (username, password) for access to the host if remote.
            defer_wmi (optional, default=True): Only valid for Windows servers. If True, defers initializing the WMI interface until first use.
            ip (optional): The server IP address. If not specified, will default to the IP of the localhost.
            os_type (optional): The server type. If not specified, will default to the type of the localhost.

        Attributes:
            _auth: The value of the auth argument.
            _domain: The derived value of the domain argument.
            _hostname: The derived value of the hostname argument.
            _ip: The derived value of the ip argument.
            _os_manager: The remote management interface for remote servers, None otherwise.
            _os_type: The value of the os_type argument.
            _wmi_manager: The WMI object.

        Raises:
            ServerObjectManagementError.SERVER_NOT_FOUND: If the remote server IP is not found.
        """
        self._hostname = (hostname if hostname else node().split('.')[0]).lower()
        try:
            self._domain = (domain if domain else getfqdn().split('.', 1)[1]).lower()
        except IndexError:
            self._domain = ''
        self._auth = auth
        self._ip = ip
        self._os_type = os_type
        self._wmi_manager: Optional[WMI] = None
        if not self._ip:
            try:
                self._ip = gethostbyname(self.fqdn)
            except gaierror as err:
                server_found = False
                if err.errno not in (self._WSAHOST_NOT_FOUND, self._WSA_NAME_OR_SERVICE_NOT_KNOWN):
                    raise
            else:
                server_found = True
            if not server_found:
                if self.is_local:
                    self._ip = '127.0.0.1'
                else:
                    raise ServerObjectManagementError(ServerObjectManagementError.SERVER_NOT_FOUND, server=self.fqdn)
        self._os_manager = OSManager('' if self.is_local else self.hostname, self._auth)
        if not defer_wmi:
            self._connect_wmi()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def _connect_wmi(self) -> None:
        """Make a WMI connection to the server.

        Returns:
            Nothing.

        Raises:
            ServerObjectManagementError.REMOTE_CONNECTION_ERROR: If there was an error connecting to the WMI manager.
        """
        manager_args = {} if self.is_local else {'computer': self.hostname}
        if isinstance(self._auth, tuple):
            manager_args['user'] = self._auth[0]
            manager_args['password'] = self._auth[1]

        try:
            self._wmi_manager = WMI(**manager_args)
        except x_wmi as err:
            raise ServerObjectManagementError(ServerObjectManagementError.REMOTE_CONNECTION_ERROR, server=self.hostname, msg=str(err)) from err

    def _get_object_manager(self, item_type: str, wmi: WMIObject, /) -> Optional[ServerManager]:
        """Return the correct object manager for the platform and item type.

        Args:
            item_type: The item type for which the object manager is requested.
            wmi: The WMI manager.

        Returns:
            The object manager.

        Raises:
            ServerObjectManagementError.REMOTE_NOT_SUPPORTED: If the platform is Linux or the item_type is Service.
        """
        if (item_type != 'Service') and (self.os_type != OsType.windows) and not self.is_local:
            raise ServerObjectManagementError(ServerObjectManagementError.REMOTE_NOT_SUPPORTED)
        if wmi and not self._wmi_manager:
            self._connect_wmi()
        return self._wmi_manager if wmi else self._os_manager

    domain = property(lambda s: s._domain, doc='A read-only property which returns the domain of the server.')
    fqdn = property(lambda s: f'{s.hostname}.{s.domain}' if s.domain else s.hostname, doc='A read-only property which returns the full-qualified domain name of the server.')
    hostname = property(lambda s: s._hostname, doc='A read-only property which returns the hostname of the server.')
    ip = property(lambda s: s._ip, doc='A read-only property which returns IP for the server.')
    is_local = property(lambda s: getfqdn().lower() == s.fqdn, doc='A read-only property which returns True if the server is the local host.')
    os_type = property(lambda s: s._os_type, doc='A read-only property which returns the OS type of the server.')

    def create_management_object(self, item_type: str, unique_id: str, wmi: WMIObject = _DEFAULT_WMI, /, *, error_if_exists: bool = True, **key_args) -> 'ManagementObject':
        """Create a management object of the specified type.

        Args:
            item_type: The item type of the object to create.
            unique_id: The unique object identifier.
            wmi (optional): True if this is a Windows platform, False otherwise.
            error_if_exists (optional, default=True): If True raise an error if the object already exists.
            **key_args (optional, default={}): A dictionary of key value pairs to pass to add to the create object.

        Returns:
            The created management object.

        Raises:
            ServerObjectManagementError.WMI_ERROR: If there was an error creating the object.
        """
        manager = self._get_object_manager(item_type, wmi)
        unique_key = 'ProcessId' if (item_type == 'Process') else 'Name'
        creation_args = {unique_key: unique_id, 'DisplayName': unique_id}

        created_object = self.get_unique_object(item_type, wmi, **{unique_key: unique_id})
        result = getattr(manager, ManagementObject.OBJECT_PREFIX + item_type).Create(**creation_args, **key_args)[0] \
            if error_if_exists or not created_object else False

        if result:
            msg = self._WMI_SERVICE_CREATE_ERRORS[result] if (result in self._WMI_SERVICE_CREATE_ERRORS) else f'Return value: {result}'
            raise ServerObjectManagementError(ServerObjectManagementError.WMI_ERROR, server=self.hostname, msg=msg)

        if not created_object:
            created_object = self.get_unique_object(item_type, wmi, **{unique_key: unique_id})
        return globals()[item_type](created_object, manager, unique_key, unique_id)

    def create_scheduled_task(self, task: str, /, *, exe: str, schedule_type: str, schedule: str, user: str = '', password: str = '',  # pylint: disable=too-many-locals
                              start_in: Optional[PathName] = None, disable: bool = False) -> 'ScheduledTask':
        """Create a scheduled task.

        Args:
            task: The name of the scheduled task to create.
            exe: The executable the scheduled task will run.
            schedule_type: The schedule type.
            schedule: The schedule on which to run the task.
            user (optional, default=''): In not None, the user account that the task will run under.
            password (optional, default=''): The password for the user account that the task will run under.
            start_in (optional, default=None): If not None, the directory in which the task will start.
            disable (optional, default=False): If True the task will be created as disabled.

        Returns:
            The created scheduled task.
        """
        if isinstance(exe, str):
            exe_args: List[str] = []
        else:
            exe_args = exe[1:]
            exe = exe[0]

        cmd_args = (create_base := ['/Create', '/F', '/TN', task]) + ['/TR', ' '.join([f'"{exe}"'] + exe_args),
                                                                      '/SC', schedule_type,
                                                                      ('/ST' if (schedule_type.lower() == 'daily') else '/MO'), schedule]

        if user:
            cmd_args += ['/RU', user]
        if password:
            cmd_args += ['/RP', password]
        if not self.is_local:
            cmd_args += ['/S', self.hostname]
        if isinstance(self._auth, tuple):
            cmd_args += ['/U', self._auth[0], '/P', self._auth[1]]
        _run_task_scheduler(*cmd_args)
        task_object = self.get_scheduled_task(task)

        if disable:
            task_object.manage(TaskSignal.disable)

        if start_in is None:
            start_in = Path(exe).parent
        if start_in:
            task_xml = xml_parse(task_file := ScheduledTask.TASK_HOME / task)
            exec_el = cast(Element, task_xml.find('ns:Actions/ns:Exec', {'ns': ScheduledTask.TASK_NAMESPACE}))
            SubElement(exec_el, f'{{{ScheduledTask.TASK_NAMESPACE}}}WorkingDirectory').text = str(start_in)
            task_xml.write(task_file)
            cmd_args = create_base + ['/XML', str(task_file)]
            if password:
                cmd_args += ['/RU', user, '/RP', password]
            _run_task_scheduler(*cmd_args)

        return task_object

    def create_service(self, service: str, /, *, exe: str, user: str = '', password: str = '', start: bool = False,
                       timeout: int = 0, error_if_exists: bool = True) -> 'Service':
        """Create a service.

        Args:
            service: The name of the service to create.
            exe: The executable the service will run.
            user (optional, default=''): In not None, the user account that the service will run under.
            password (optional, default=''): The password for the user account that the service will run under.
            timeout (optional, default=0): If not 0, the number of seconds to wait for the service to start.
            error_if_exists (optional, default=True): If True raise an error if the object already exists.

        Returns:
            The created service.
        """
        service_obj = cast(Service, self.create_management_object('Service', service, PathName=str(exe), StartName=user, StartPassword=password, error_if_exists=error_if_exists))
        if start:
            service_obj.manage(ServiceSignal.start, timeout=timeout)
        return service_obj

    def get_management_objects(self, item_type: str, wmi: WMIObject = _DEFAULT_WMI, /, **filters) -> List['NamedOSObject']:
        """Get a list of management objects.

        Args:
            item_type: The item type of the objects.
            wmi (optional): True if this is a Windows platform, False otherwise.
            **filters (optional, default={}): A dictionary of filters to pass to the manager to filter the objects returned.

        Returns:
            The list of management objects.
        """
        manager = self._get_object_manager(item_type, wmi)
        unique_key = 'ProcessId' if (item_type == 'Process') else 'Name'
        extra_keys = {}
        for (key, item_list) in {'service_type': 'Service'}.items():
            if (item_type in item_list) and (key in filters):
                if self.os_type != OsType.windows:
                    extra_keys[key] = filters[key]
                else:
                    del filters[key]
        return [globals()[item_type](r, manager, unique_key, getattr(r, unique_key), **extra_keys) for r in getattr(manager, ManagementObject.OBJECT_PREFIX + item_type)(**filters)]

    def get_object_by_name(self, item_type: str, name: str, wmi: bool = _DEFAULT_WMI, /, **filters) -> Optional['NamedOSObject']:
        """Get a management object by name.

        Args:
            item_type: The item type of the object.
            name: The name of the object.
            wmi (optional): True if this is a Windows platform, False otherwise.
            **filters (optional, default={}): A dictionary of filters to pass to the manager to filter for the object.

        Returns:
            The requested management object.

        Raises:
            ServerObjectManagementError.NOT_UNIQUE: If more than one management object was found.
        """
        return self.get_unique_object(item_type, wmi, Name=name, **filters)

    def get_path(self, the_path: str, /) -> ServerPath:
        """Get the specified path as a ServerPath object.

        Args:
            the_path: The path to return.

        Return:
            The specified path as a ServerPath object.
        """
        return ServerPath(self, the_path)

    def get_process_list(self, **filters) -> List['Process']:
        """Get the process list for this server.

        Returns:
            The process list for this server.
        """
        return cast(List['Process'], self.get_management_objects('Process', **filters))

    def get_scheduled_task(self, task: str, /) -> 'ScheduledTask':
        """Get the specified scheduled task.

        Returns:
            The specified scheduled task.
        """
        return cast('ScheduledTask', self.get_object_by_name('ScheduledTask', task, False))

    def get_scheduled_task_list(self) -> List['ScheduledTask']:
        """Get the scheduled tasks for this server.

        Returns:
            The scheduled tasks for this server.
        """
        cmd_args = ['/Query', '/FO', 'CSV']
        if not self.is_local:
            cmd_args += ['/S', self.fqdn]
        if isinstance(self._auth, tuple):
            cmd_args += ('/U', self._auth[0], '/P', self._auth[1])
        return [self.get_scheduled_task(t['TaskName']) for t in DictReader(_run_task_scheduler(*cmd_args))]

    def get_service(self, service: str, /, **key_args) -> 'Service':
        """Get the specified service.

        Args:
            **key_args (optional, default={}): A dictionary of key value pairs to apply as filters when retrieving the service.

        Returns:
            The specified service.
        """
        if 'service_type' not in key_args:
            if self.os_type == OsType.windows:
                service_type = ServiceType.windows
            elif self.get_path('/sbin/initctl').exists():
                service_type = ServiceType.upstart
            elif self.get_path('/bin/systemctl').exists():
                service_type = ServiceType.systemd
            else:
                service_type = ServiceType.sysv
            key_args['service_type'] = service_type
        return cast('Service', self.get_object_by_name('Service', service, **key_args))

    def get_service_list(self, service_type: Optional[ServiceType] = None, /) -> List['Service']:
        """Get the services for this server.

        Args:
            service_type (optional, default=None): If None, determine the service type based on the OS.

        Returns:
            The services for this server.
        """
        if service_type is None:
            if self.os_type == OsType.windows:
                service_type = ServiceType.windows
            elif self.get_path('/sbin/initctl').exists():
                service_type = ServiceType.upstart
            elif self.get_path('/bin/systemctl').exists():
                service_type = ServiceType.systemd
            else:
                service_type = ServiceType.sysv
        return cast(List['Service'], self.get_management_objects('Service', service_type=service_type))

    def get_unique_object(self, item_type: str, wmi: WMIObject = _DEFAULT_WMI, /, **filters) -> Optional['NamedOSObject']:
        """Get the requested management object which must be unique.

        Args:
            item_type: The item type of the object.
            wmi (optional): True if this is a Windows platform, False otherwise.
            **filters (optional, default={}): A dictionary of filters to pass to the manager to filter for the object.

        Returns:
            The requested management object.

        Raises:
            ServerObjectManagementError.NOT_UNIQUE: If more than one management object was found.
        """
        match len(results := self.get_management_objects(item_type, wmi, **filters)):
            case 0:
                return None
            case 1:
                return results[0]
        raise ServerObjectManagementError(ServerObjectManagementError.NOT_UNIQUE, type=item_type, filters=filters)

    def remove_management_object(self, item_type: str, unique_id: str, wmi: WMIObject = _DEFAULT_WMI, /, *, error_if_not_exists: bool = False) -> None:
        """Remove a management object.

        Args:
            item_type: The item type of the object to remove.
            unique_id: The unique identifier of the object to remove.
            wmi (optional): True if this is a Windows platform, False otherwise.
            error_if_not_exists (optional, default=False): If False raise an error if the object does not exist.

        Returns:
            Nothing.

        Raises:
            ServerObjectManagementError.WMI_ERROR: If there was an error removing the object.
        """
        unique_key = 'ProcessId' if (item_type == 'Process') else 'Name'
        removal_object = cast('ManagementObject', self.get_unique_object(item_type, wmi, **{unique_key: unique_id}))
        if error_if_not_exists and not removal_object:
            raise ServerObjectManagementError(ServerObjectManagementError.OBJECT_NOT_FOUND, type=item_type)
        if (result := removal_object.Delete()[0] if removal_object else False):
            msg = self._WMI_SERVICE_CREATE_ERRORS[result] if (result in self._WMI_SERVICE_CREATE_ERRORS) else f'Return value: {result}'
            raise ServerObjectManagementError(ServerObjectManagementError.WMI_ERROR, server=self.hostname, msg=msg)

    def remove_scheduled_task(self, task: str, /) -> None:
        """Remove the specified scheduled task.

        Args:
            task: The name of the scheduled task to remove.

        Returns:
            Nothing.
        """
        self.get_scheduled_task(task).remove()

    def remove_service(self, service: str, /, *, error_if_not_exists: bool = False) -> None:
        """Remove the specified service.

        Args:
            service: The name of the service to remove.
            error_if_not_exists (optional, default=False): If False raise an error if the object does not exist.

        Returns:
            Nothing.
        """
        self.remove_management_object('Service', service, error_if_not_exists=error_if_not_exists)

    def run_command(self, command: str, /, *cmd_args, **sys_cmd_args) -> CommandResult:
        """Run a command on the server.

        Args:
            command: The command to run.
            *cmd_args (optional, default=[]): The arguments to pass to the command.
            *sys_cmd_args (optional, default={}): A dictionary of named arguments to pass to the command.

        Returns:
            The result of the command.
        """
        if (not self.is_local) and self._auth:
            sys_cmd_args['remote_auth'] = self._auth
        return syscmd(command, *cmd_args, remote=(False if self.is_local else self.ip), remote_is_windows=(not self.is_local) and (self.os_type == OsType.windows), **sys_cmd_args)

    if sys.platform == 'win32':
        def get_iis_instance(self) -> IISInstance:
            """Get the IIS instance for this server.

            Returns:
                The IIS instance for this server.
            """
            return IISInstance(self.fqdn if not self.is_local else None)

        def get_process_connection(self, process: str, /) -> 'COMObject':
            """Get a COM connection to the specified process.

            Returns:
                The COM connection to the specified process.
            """
            return COMObject(process, self.ip)


class OSManager:
    """Class to make non WMI OS management look like WMI management."""

    def __init__(self, computer: str = '', auth: ServerAuthType = None):
        """
        Args:
            computer (optional, default=''): The remote computer.
            auth (optional, default=None): A (username, password) tuple for remote server access.

        Attributes:
            auth: The value of the auth argument.
            computer: The value of the computer argument.
        """
        self.computer = computer
        self.auth = auth

    def get_object_as_list(self, object_type: str, /, *, Name: str, **key_args) -> List['ManagementObject']:
        """Get the specified OS object.

        Args:
            object_type: The object type.
            Name: The name of the object to return.
            **key_args (optional, default={}): A dictionary of filters to pass to the manager to filter the objects returned.

        Returns:
            The list of management objects.
        """
        try:
            return [globals()[object_type](Name, self.computer, self.auth, **key_args)]
        except ServerObjectManagementError as err:
            if err.code == ServerObjectManagementError.OBJECT_NOT_FOUND.code:
                return []
            raise

    def LinuxProcess(self, CommandLine: Optional[str] = None, ExecutablePath: Optional[str] = None,
                     Name: Optional[str] = None, ProcessId: Optional[int] = None) -> List['LinuxProcess']:
        """Get the specified Linux process.

        Args:
            Exactly one of these options must be specified:
            CommandLine (optional, default=None): The command line for the process.
            ExecutablePath (optional, default=None): The command line for the process.
            Name (optional, default=None): The command line for the process.
            ProcessId (optional, default=None): The command line for the process.

        Returns:
            The specified Linux process.

        Raises:
            ServerObjectManagementError.BAD_FILTER: If more than one selection option is specified.
        """
        if len([v for v in (CommandLine, ExecutablePath, Name, ProcessId) if v is not None]) != 1:
            raise ServerObjectManagementError(ServerObjectManagementError.BAD_FILTER)

        if ProcessId:
            try:
                return [LinuxProcess(ProcessId)]
            except NoSuchProcess:
                return []

        process_list = [p for p in process_iter(attrs=('pid', 'cmdline', 'exe', 'name'))]  # pylint: disable=unnecessary-comprehension
        if CommandLine:
            return [LinuxProcess(p.pid) for p in process_list if p.info['cmdline'] == CommandLine]  # type: ignore[attr-defined]
        if ExecutablePath:
            return [LinuxProcess(p.pid) for p in process_list if p.info['exe'] == ExecutablePath]  # type: ignore[attr-defined]
        if Name:
            return [LinuxProcess(p.pid) for p in process_list if p.info['name'] == Name]  # type: ignore[attr-defined]
        return []

    def LinuxService(self, Name: str, service_type: ServiceType, /) -> 'LinuxService':
        """Get the specified Linux service.

        Args:
            Name: The name of the service to return.
            service_type: The type of service to return.

        Returns:
            The specified Linux service.
        """
        return cast('LinuxService', self.get_object_as_list('LinuxService', Name=Name, service_type=service_type))

    def Win32_ScheduledTask(self, Name: str, /) -> 'ScheduledTask':
        """Get the specified Windows Scheduled Task.

        Args:
            Name: The name of the scheduled task to return.

        Returns:
            The specified Windows scheduled task.
        """
        return cast('ScheduledTask', self.get_object_as_list('Win32_ScheduledTask', Name=Name))


class NamedOSObject:  # pylint: disable=too-few-public-methods
    """Class to allow management of all OS objects using a similar interface."""

    def __init__(self, Name: str, computer: str, auth: ServerAuthType, /):
        """
        Args:
            Name: The name of the object.
            computer: The remote computer.
            auth: A (username, password) tuple for remote server access.

        Attributes:
            auth: The value of the auth argument.
            computer: The value of the computer argument.
            Name: The value of the Name argument.
        """
        self.Name = Name
        self.computer = computer
        self.auth = auth
        self.validate()  # type: ignore[attr-defined]  # pylint: disable=no-member


class LinuxService(NamedOSObject):
    """Class to create a universal abstract interface for a Linux daemon service."""

    def __init__(self, Name: str, computer: str, auth: ServerAuthType, service_type: ServiceType, /):
        """
        Args:
            Name: The name of the object.
            computer: The remote computer.
            auth: A (username, password) tuple for remote server access.
            service_type: The type of the Linux service

        Attributes:
            type: The value of the service_type argument.
        """
        self.type = service_type
        super().__init__(Name, computer, auth)

    def _manage(self, command: str, /) -> CommandResult | CMDError:
        """Manage the service.

        Args:
            command: The management action to perform.

        Returns:
            The result of the management command.
        """
        control_command = ['service', self.Name, command] if (self.type == ServiceType.sysv) \
            else ['systemctl' if (self.type == ServiceType.systemd) else 'initctl', command, self.Name]
        try:
            return syscmd(*control_command, use_shell=True, ignore_stderr=True, remote=self.computer, remote_auth=self.auth)
        except CMDError as err:
            if command == 'status':
                return err
            raise

    @property
    def state(self) -> str:
        """A read-only property which returns the state value of the service."""
        if (result := self._manage('status')) and isinstance(result, list) and ('stop' in result[0]):  # pylint: disable=used-before-assignment
            return 'Stopped'
        if not hasattr(result, 'vars'):
            return 'Running'
        if (error := cast(CMDError, result)).vars['returncode'] == 3:
            return 'Stopped'
        raise error

    def DisableService(self) -> None:
        """Disable the service.

        Returns:
            Nothing.
        """
        self.StopService()
        self._manage('disable')

    def EnableService(self) -> None:
        """Enable the service.

        Returns:
            Nothing.
        """
        self._manage('enable')
        self.StartService()

    def RestartService(self) -> None:
        """Restart the service.

        Returns:
            Nothing.
        """
        self._manage('restart')

    def StartService(self) -> None:
        """Start the service.

        Returns:
            Nothing.
        """
        self._manage('start')

    def StopService(self) -> None:
        """Stop the service.

        Returns:
            Nothing.
        """
        self._manage('stop')

    def validate(self) -> None:
        """Determine if the service exists.

        Returns:
            Nothing.

        Raises:
            ServerObjectManagementError.OBJECT_NOT_FOUND: If the service is not found.
        """
        try:
            self.state
        except CMDError as err:
            if not err.vars['returncode'] == (4 if (self.type == ServiceType.systemd) else 1):
                raise
            raise ServerObjectManagementError(ServerObjectManagementError.OBJECT_NOT_FOUND, type=type(self).__name__) from err


class LinuxProcess:
    """Class to create a universal abstract interface for a Linux process."""

    def __init__(self, ProcessId: int, /):
        """
        Args:
            ProcessId: The process ID of the process.

        Attributes:
            process_obj: The API object representing the process.
            ProcessId: The value of the ProcessId argument.
        """
        self.ProcessId = ProcessId
        self.process_obj = _LinuxProcess(self.ProcessId)

    CommandLine = property(lambda s: s.process_obj.cmdline(), doc='A read-only property which returns the command line for the process.')
    ExecutablePath = property(lambda s: s.process_obj.exe(), doc='A read-only property which returns the executable path for the process.')
    Name = property(lambda s: s.process_obj.name(), doc='A read-only property which returns the name of the process.')

    def Kill(self) -> None:
        """Kill the process.

        Returns:
            Nothing.
        """
        self.process_obj.kill()

    def Terminate(self) -> None:
        """Terminate the process.

        Returns:
            Nothing.
        """
        self.process_obj.terminate()


class LinuxScheduledTask:  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Linux cron job."""


class Win32_ScheduledTask(NamedOSObject):
    """Class to abstract a Windows Scheduled Task since they are not available using WMI."""

    def __getattr__(self, attr: str):
        if attr == 'state':
            attr = 'scheduled_task_state'
        attr = attr.title().replace('_', ' ')
        task_info = [line for line in DictReader(self._run_task_scheduler('/Query', '/V', '/FO', 'CSV'))][0]  # pylint: disable=unnecessary-comprehension
        if attr not in task_info:
            raise AttributeError(f"'{type(self)}' object has no attribute '{attr}'")
        return task_info[attr]

    def _run_task_scheduler(self, *cmd_args, **sys_cmd_args) -> CommandResult:
        """Run the Windows sc command to manage the service.

        Args:
            *cmd_args (optional, default=[]): The command and arguments.
            **sys_cmd_args (optional, default={}): The named arguments.

        Returns:
            The result of the sc command.
        """
        args = list(cmd_args) + ['/TN', self.Name]
        if self.computer:
            args += ['/S', self.computer]
        if isinstance(self.auth, tuple):
            args += ('/U', self.auth[0], '/P', self.auth[1])
        return _run_task_scheduler(*args, **sys_cmd_args)

    def manage(self, signal: TaskSignal, /, *, wait: bool = True) -> None:
        """Manage the scheduled task.

        Args:
            signal: The signal to send to the scheduled task.
            wait (optional, default=True): If True, wait for the action to complete.

        Returns:
            Nothing.

        Raises:
            ServerObjectManagementError.BAD_OBJECT_SIGNAL: If the signal is unknown.
        """
        control_args: Tuple = tuple()
        match signal:
            case TaskSignal.enable:
                control_args = ('/Change', '/ENABLE')
            case TaskSignal.disable:
                control_args = ('/Change', '/DISABLE')
            case TaskSignal.run:
                control_args = ('/Run',)
            case TaskSignal.end:
                control_args = ('/End',)
            case _:
                raise ServerObjectManagementError(ServerObjectManagementError.BAD_OBJECT_SIGNAL, signal=signal.name)

        self._run_task_scheduler(*control_args)
        while wait and (self.status.lower() == 'running'):
            sleep(_STATUS_CHECK_INTERVAL)

    def remove(self) -> CommandResult:
        """Remove the scheduled task.

        Returns:
            The result of the remove command.
        """
        return self._run_task_scheduler('/Delete', '/F')

    def validate(self) -> None:
        """Determine if the scheduled task exists.

        Returns:
            Nothing.

        Raises:
            ServerObjectManagementError.OBJECT_NOT_FOUND: If the scheduled task is not found.
        """
        try:
            self.state
        except CMDError as err:
            if not err.vars['returncode'] == 1:
                raise
            raise ServerObjectManagementError(ServerObjectManagementError.OBJECT_NOT_FOUND, type=type(self).__name__) from err


class ManagementObject:
    """Management object to provide OS independent interface.

    Attributes:
        OBJECT_PREFIX: The prefix to use to correctly translate to a NamesOSObject type.
    """
    OBJECT_PREFIX = 'Win32_' if WIN32 else 'Linux'

    def __init__(self, object_ref: 'ManagementObject', manager: OSManager, key: str, value: str, /, **key_values):
        """
        Args:
            object_ref: A reference to the management object.
            manager: A reference to the manager of the object.
            key: The unique key used to identify the object to the manager.
            value: The unique value used to identify the object to the manager.
            key_values (optional): A additional dictionary of key/value pairs to apply when referencing the object in the manager.

        Attributes:
            key: The value of the key argument.
            key_values: The value of the key_values argument.
            manager: The value of the manager argument.
            object_ref: The value of the object_ref argument.
            type: The object type.
            value: The value of the value argument.
        """
        self.type = self.OBJECT_PREFIX + type(self).__name__
        self.object_ref: Optional[ManagementObject] = object_ref
        self.manager = manager
        self.key = key
        self.value = value
        self.key_values = key_values

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        del self.object_ref
        return False

    def __getattr__(self, attr: str):
        self.refresh()
        return getattr(self.object_ref, attr)

    def refresh(self) -> None:
        """Refresh the state of the object.

        Returns:
            Nothing.

        Raises:
            ServerObjectManagementError.NOT_UNIQUE: If the object is not unique.
        """
        if self.object_ref:
            if len(results := cast(List[ManagementObject], getattr(self.manager, self.type)(**{self.key: self.value}, **self.key_values))) > 1:
                raise ServerObjectManagementError(ServerObjectManagementError.NOT_UNIQUE, type=self.type, key=self.key, val=self.value)
            self.object_ref = results[0] if results else None


class Service(ManagementObject):
    """Class to create a universal abstract interface for an OS service."""

    def __getattr__(self, attr: str):
        if attr == 'state':
            return self.ServiceState[super().__getattr__(attr).replace(' ', '')]
        return super().__getattr__(attr)

    def manage(self, signal: ServiceSignal, /, *, wait: bool = True, ignore_state: bool = False, timeout: int = False) -> None:  # pylint: disable=too-many-branches,too-many-statements
        """Manage the service.

        Args:
            signal: The signal to send to the service.
            wait (optional, default=True): If True, wait for the action to complete.
            ignore_state (optional, default=False): If False, do not check the initial state before performing the action.
            timeout (optional, default=0): If wait is True, the number of seconds to wait for the action to complete, indefinitely if 0.

        Returns:
            Nothing.

        Raises:
            ServerObjectManagementError.BAD_OBJECT_SIGNAL: If the requested action is invalid.
            ServerObjectManagementError.BAD_OBJECT_STATE: If ignore_state is False and the current state is invalid.
            ServerObjectManagementError.BAD_TRANSITION: If ignore_state is False and the requested action is not valid for the current state.
            ServerObjectManagementError.STATUS_CHECK_TIMEOUT: If wait is True and timeout is not False and the object has not reached the required state.
        """
        match signal:
            case self.ServiceSignal.enable | self.ServiceSignal.start | self.ServiceSignal.resume | self.ServiceSignal.restart:
                match signal:
                    case self.ServiceSignal.enable:
                        control_method = 'EnableService'
                    case self.ServiceSignal.start:
                        control_method = 'StartService'
                    case self.ServiceSignal.resume:
                        control_method = 'ResumeService'
                    case self.ServiceSignal.restart:
                        control_method = 'RestartService'
                final_state = self.ServiceState.Running
            case self.ServiceSignal.disable | self.ServiceSignal.stop:
                control_method = 'StopService' if self.ServiceSignal.stop else 'DisableService'
                final_state = self.ServiceState.Stopped
            case self.ServiceSignal.pause:
                control_method = 'PauseService'
                final_state = self.ServiceState.Paused
            case _:
                raise ServerObjectManagementError(ServerObjectManagementError.BAD_OBJECT_SIGNAL, signal=signal.name)

        control_method = ''
        if not ignore_state:
            wait_for = self.ServiceState.Running
            match self.state:
                case self.ServiceState.StartPending | self.ServiceState.Running | self.ServiceState.ContinuePending:
                    wait_for = self.ServiceState.Running
                    if signal in (self.ServiceSignal.enable, self.ServiceSignal.start, self.ServiceSignal.resume):
                        control_method = ''
                case self.ServiceState.StopPending | self.ServiceState.Stopped:
                    wait_for = self.ServiceState.Stopped
                    match signal:
                        case self.ServiceSignal.disable, self.ServiceSignal.stop:
                            control_method = ''
                        case self.ServiceSignal.resume, self.ServiceSignal.restart:
                            control_method = 'StartService'
                        case self.ServiceSignal.pause:
                            raise ServerObjectManagementError(ServerObjectManagementError.BAD_TRANSITION, from_state='stopped', to_state='paused')
                case self.ServiceState.PausePending | self.ServiceState.Paused:
                    wait_for = self.ServiceState.Paused
                    if signal == self.ServiceSignal.start:
                        control_method = 'ResumeService'
                case _:
                    raise ServerObjectManagementError(ServerObjectManagementError.BAD_OBJECT_STATE, state=self.state)
            wait_length = 0
            while self.state != wait_for:
                sleep(_STATUS_CHECK_INTERVAL)
                wait_length += _STATUS_CHECK_INTERVAL
                if timeout and (wait_length > timeout):
                    raise ServerObjectManagementError(ServerObjectManagementError.STATUS_CHECK_TIMEOUT, type='service', state=wait_for.name)

        if control_method:
            if WIN32 and control_method == 'RestartService':
                self.manage(self.ServiceSignal.stop, wait=wait, ignore_state=ignore_state)
                self.manage(self.ServiceSignal.start, wait=wait, ignore_state=ignore_state)
            else:
                getattr(self, control_method)()
                sleep(_STATUS_CHECK_INTERVAL)
        wait_length = 0
        while wait and (self.state != final_state):
            sleep(_STATUS_CHECK_INTERVAL)
            wait_length += _STATUS_CHECK_INTERVAL
            if timeout and (wait_length > timeout):
                raise ServerObjectManagementError(ServerObjectManagementError.STATUS_CHECK_TIMEOUT, type='service', state=final_state.name)


class Process(ManagementObject):
    """Class to create a universal abstract interface for an OS process."""

    def manage(self, signal: ProcessSignal, /, *, wait: bool = True, timeout: bool = False) -> None:
        """Manage the process.

        Args:
            signal: The signal to send to the process.
            wait (optional, default=True): If True, wait for the action to complete.
            timeout (optional, default=False): If wait is True, the number of seconds to wait for the action to complete, indefinitely if False.

        Returns:
            Nothing.

        Raises:
            ServerObjectManagementError.BAD_OBJECT_SIGNAL: If the requested action is invalid.
            ServerObjectManagementError.STATUS_CHECK_TIMEOUT: If wait is True and timeout is not False and the object has not reached the required state.
        """
        control_method = ''
        match signal:
            case self.ProcessSignal.stop:
                control_method = 'Terminate'
            case self.ProcessSignal.kill:
                control_method = 'Terminate' if WIN32 else 'Kill'
            case _:
                raise ServerObjectManagementError(ServerObjectManagementError.BAD_OBJECT_SIGNAL, signal=signal.name)

        if control_method:
            getattr(self, control_method)()
        wait_length = 0
        while wait and self.object_ref:
            sleep(_STATUS_CHECK_INTERVAL)
            self.refresh()
            wait_length += _STATUS_CHECK_INTERVAL
            if timeout and (wait_length > timeout):
                raise ServerObjectManagementError(ServerObjectManagementError.STATUS_CHECK_TIMEOUT, type='process', state=signal.name)


class ScheduledTask(ManagementObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for an OS scheduled task.

    Attributes:
        TASK_HOME: The directory where task definitions are stored.
        TASK_NAMESPACE: The Windows task namespace needed to parse the XML task definitions.
    """
    TASK_HOME = Path(environ['SystemRoot'], 'system32/Tasks') if WIN32 else Path('/opt/cronjobs')
    TASK_NAMESPACE = 'http://schemas.microsoft.com/windows/2004/02/mit/task'


if sys.platform == 'win32':
    class COMObject:
        """Class to create a universal abstract interface for a Windows COM object."""

        def __init__(self, ref: Union[str, 'COMObject'], hostname: str = '', /):
            """
            Args:
                ref: The Windows COM object.
                hostname (optional): The server hostname. If not specified, will default to the name of the localhost.

            Attributes:
                _connection: The value of the ref argument.
                _hostname: The derived value of the hostname argument.

            Raises:
                ServerObjectManagementError.REMOTE_CONNECTION_ERROR: If there was a failure making a connection to the COM object.
            """
            self._hostname = hostname
            try:
                if isinstance(ref, CDispatch):
                    self._connection = ref
                    str(self._connection)
                else:
                    self._connection = DispatchEx(ref, self._hostname)
            except com_error as err:
                raise ServerObjectManagementError(ServerObjectManagementError.REMOTE_CONNECTION_ERROR, server=self._hostname, msg=str(err)) from err

        def __enter__(self):
            return self

        def __exit__(self, *exc_info):
            self.disconnect()
            return False

        def __getattr__(self, attr: str):
            return getattr(self._connection, attr)

        def __setattr__(self, attr: str, value: str):
            if attr.startswith('_'):
                super().__setattr__(attr, value)
                return
            setattr(self._connection, attr, value)

        def disconnect(self) -> None:
            """Disconnect the COM object.

            Returns:
                Nothing.
            """
            del self._connection


def get_server_object(server: ServerType, /) -> Server:
    """Get a server object for the specific fqdn.

    Args:
        server: The fqdn string for the server

    Returns:
        Returns the Server object.
    """
    if isinstance(server, Server):
        return server
    if not isinstance(server, str):
        raise TypeError(server)
    return Server(*(server.split('.', 1)))


def _run_task_scheduler(*cmd_args, **sys_cmd_args) -> CommandResult:
    """Interface to run the standard Windows schtasks command-line tool.

    Args:
        *cmd_args: The arguments to pass to schtasks.
        **sys_cmd_args: The arguments to pass to syscmd when running schtasks.

    Returns:
        The result of the syscmd call to schtasks.
    """
    return syscmd('schtasks.exe', *cmd_args, **sys_cmd_args)

# cSpell:ignore cmdline sbin wsahost psutil syscmd iispy platarch serverpath schtasks



"""Unit tests for the fileutil module."""

# pylint: disable=missing-class-docstring,missing-function-docstring,invalid-name
# flake8: noqa

from datetime import datetime as dt, timedelta
from os import utime
from pathlib import Path
from stat import S_IREAD
from tempfile import mkdtemp
from time import mktime, time
from unittest import main, TestCase

from batcave.fileutil import prune
from batcave.sysutil import rmtree_hard


class TestPrune(TestCase):
    @property
    def _file_list(self):
        return sorted(list(self._tempdir.iterdir()))

    def _prune(self, age, **kwargs):
        prune(self._tempdir, age, **kwargs)

    def _print_list(self, full=False):
        print([f.name for f in (self._full_file_list if full else self._file_list)])

    def setUp(self):
        self._tempdir = Path(mkdtemp()).resolve()
        self._time = time()
        file_time = dt.now()
        for c in 'ab':
            for i in (1, 2):
                Path(temp_file1 := self._tempdir / f'{c}{i}.Ext{i}').touch()
                Path(temp_file2 := self._tempdir / f'{c}{i}.Txt{i}').touch()
                if i == 1:
                    file_time += timedelta(days=2)
                utime(temp_file1, (file_epoch_time := mktime(file_time.timetuple()), file_epoch_time))
                utime(temp_file2, (file_epoch_time, file_epoch_time))
        self._full_file_list = self._file_list

    def tearDown(self):
        rmtree_hard(self._tempdir)

    def test_prune_1_no_removal(self):
        self._prune(10)
        self.assertEqual(self._full_file_list, self._file_list)

    def test_prune_2_simple(self):
        self._prune(age=10)
        self.assertEqual(self._full_file_list, self._file_list)
        self._prune(age=2)
        self.assertEqual(self._full_file_list[:-4], self._file_list)

    def test_prune_3_by_ext(self):
        self._prune(age=2, exts=['.txt'])
        self.assertEqual(self._full_file_list, self._file_list)
        self._prune(age=2, exts=['.Txt2'])
        self.assertEqual(self._full_file_list[:-1], self._file_list)

    def test_prune_4_by_ext_ignore_case(self):
        self._prune(age=2, exts=['.txt2'], ignore_case=True)
        self.assertEqual(self._full_file_list[:-1], self._file_list)

    def test_prune_5_force(self):
        for item in self._full_file_list:
            item.chmod(S_IREAD)
        self._tempdir.chmod(S_IREAD)
        try:
            self._prune(age=2)
        except PermissionError:
            pass
        self.assertEqual(self._full_file_list, self._file_list)
        self._prune(age=2, force=True)
        self.assertEqual(self._full_file_list[:-4], self._file_list)

if __name__ == '__main__':
    main()

# cSpell:ignore batcave fileutil



"""Unit tests for the lang module."""

# pylint: disable=missing-class-docstring,missing-function-docstring,invalid-name
# flake8: noqa

import os
from unittest import main, TestCase

from batcave.lang import is_debug


class TestIsDebug(TestCase):
    def setUp(self):
        self._keeper = None
        if 'BATCAVE_DEBUG' in os.environ:
            self._keeper = os.environ['BATCAVE_DEBUG']
            del os.environ['BATCAVE_DEBUG']

    def tearDown(self):
        if self._keeper:
            os.environ['BATCAVE_DEBUG'] = self._keeper

    def test_is_debug_1_False(self):
        self.assertFalse(is_debug())

    def test_is_debug_2_True(self):
        os.environ['BATCAVE_DEBUG'] = '1'
        self.assertTrue(is_debug())

    def test_is_debug_3_SingleValue(self):
        os.environ['BATCAVE_DEBUG'] = 'TestValue'
        self.assertTrue(is_debug('TestValue'))
        self.assertFalse(is_debug('TestBadValue'))

    def test_is_debug_4_MultiValue(self):
        os.environ['BATCAVE_DEBUG'] = 'TestValue1:TestValue2'
        self.assertTrue(is_debug('TestValue1'))
        self.assertTrue(is_debug('TestValue2'))
        self.assertFalse(is_debug('TestBadValue'))


if __name__ == '__main__':
    main()

# cSpell:ignore batcave



"""This module provides utilities for managing configurations."""

# Import standard modules
from pathlib import Path
from string import Template
from typing import Optional
from xml.etree.ElementTree import ParseError

# Import internal modules
from .data import DataError, DataSource, SourceType
from .lang import BatCaveError, BatCaveException, PathName


class ConfigurationError(BatCaveException):
    """Configuration Exceptions.

    Attributes:
        BAD_FORMAT: The configuration file format is invalid.
        BAD_SCHEMA: The configuration schema is not supported.
        CONFIG_NOT_FOUND: The specified configuration file was not found.
    """
    BAD_FORMAT = BatCaveError(1, Template('Bad format for configuration file: $file'))
    BAD_SCHEMA = BatCaveError(2, Template('Invalid schema in configuration file: $file'))
    CONFIG_NOT_FOUND = BatCaveError(3, Template('Unable to find the configuration file: $file'))


class ConfigCollection:
    """This is a container class to hold a collection of configurations read from a file.

    Attributes:
        INCLUDE_CONFIG_TAG: The configuration tag which indicates an include file.
        _CURRENT_CONFIG_SCHEMA: The default DataSource schema to use.
        _MASK_MISSING: The parameter to indicate that child configurations should be ignored.
        _PARAMS_CONFIGURATION: The configuration section which is used as configuration parameters for the collection.
        _PARENT_CONFIGURATION: The parameter to indicate the parent configuration name.
    """
    _CURRENT_CONFIG_SCHEMA = 1
    _MASK_MISSING = 'mask_missing'
    _PARAMS_CONFIGURATION = 'configuration'
    _PARENT_CONFIGURATION = 'parent'

    INCLUDE_CONFIG_TAG = 'include'

    def __init__(self, name: PathName, /, *, create: bool = False, suffix: str = '_config.xml'):
        """
        Args:
            name: The configuration collection name.
            create (optional, default=False): If True the configuration file will be created if it doesn't exist.
            suffix (optional, default=_config.xml): The suffix to add to the name to derive the configuration file name.

        Attributes:
            params: This is the list of configuration parameters read from the configuration section.
            parent: The parent configuration.
            _config_filename: The derived name of the configuration file.
            _configs: This is the collection of configurations.
            _current: This is the value used by the iterator.
            _data_source: A reference to the DataSource instance read from the configuration file.
            _mask_missing: This value is read from the configuration section and
                if True will prevent missing values from the parent from creating values.
            _name: The value of the name argument.

        Raises:
            ConfigurationError.BAD_FORMAT: The the format of the configuration file is not valid.
            ConfigurationError.BAD_SCHEMA: If the schema of the configuration file is not supported.
            ConfigurationError.CONFIG_NOT_FOUND: If the derived configuration file is not found.
        """
        self._name = (path_name := Path(name)).name
        self._config_filename = path_name.parent / (path_name.name + suffix)
        try:
            self._data_source = DataSource(SourceType.xml, self._config_filename, name=self.name, schema=self._CURRENT_CONFIG_SCHEMA, create=create)
        except DataError as err:
            match err.code:
                case DataError.FILE_OPEN.code:
                    raise ConfigurationError(ConfigurationError.CONFIG_NOT_FOUND, file=self._config_filename) from err
                case DataError.BAD_SCHEMA.code:
                    raise ConfigurationError(ConfigurationError.BAD_SCHEMA, file=self._config_filename) from err
                case _:
                    raise
        except ParseError as err:
            raise ConfigurationError(ConfigurationError.BAD_FORMAT, file=self._config_filename) from err

        self.parent: Optional[ConfigCollection] = None
        self._mask_missing = True
        self.params = getattr(self, self._PARAMS_CONFIGURATION) if hasattr(self, self._PARAMS_CONFIGURATION) else None
        if hasattr(self.params, self._PARENT_CONFIGURATION):
            self.parent = ConfigCollection(getattr(self.params, self._PARENT_CONFIGURATION))
        self._mask_missing = hasattr(self.params, self._MASK_MISSING)

        self._configs = [getattr(self, t.name) for t in self._data_source.get_tables() if t.name not in (DataSource.INFO_TABLE, self._PARAMS_CONFIGURATION)]
        config_names = [c.name for c in self._configs]
        if self.parent and not self._mask_missing:
            self._configs += [c for c in self.parent if c.name not in config_names]
        self._current = 0

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()
        return False

    def __getattr__(self, attr: str):
        if self._data_source.has_table(attr):
            parent_config = getattr(self.parent, attr) if (self.parent and hasattr(self.parent, attr)) else None
            config = Configuration(self._data_source, attr, parent=parent_config)
            if hasattr(config, self.INCLUDE_CONFIG_TAG):
                config = Configuration(self._data_source, attr, parent=parent_config, include=getattr(self, getattr(config, self.INCLUDE_CONFIG_TAG)))
            return config
        if self.parent and not self._mask_missing:
            return getattr(self.parent, attr)
        raise AttributeError(f'Unknown configuration ({attr}) in {self._config_filename}')

    def __iter__(self):
        return self

    def __next__(self):
        if self._current >= len(self._configs):
            raise StopIteration()
        self._current += 1
        return self._configs[self._current - 1]

    name = property(lambda s: s._name, doc='A read-only property which returns the name of the configuration.')

    def add(self, name: str, /) -> str:
        """Add an item to the configuration collection.

        Args:
            name: The name of the item to add.

        Returns:
            The value of the added item.
        """
        self._data_source.add_table(name).add_row()
        self._data_source.commit()
        return getattr(self, name)

    def close(self) -> None:
        """Close the underlying data source."""
        self._data_source.close()


class Configuration:
    """This is a container class to hold an individual configuration in a collection."""

    def __init__(self, config_source: DataSource, name: str, /, *, parent: Optional['Configuration'] = None, include: Optional['Configuration'] = None):
        """
        Args:
            config_source: The configuration source.
            name: The configuration name.
            parent (optional, default=None): If not None, the parent configuration.
            include (optional, default=None): If not None, the configuration to include.

        Attributes:
            _data_source: The value of the config_source argument.
            _data_table: The DataTable holding the configuration values for this configuration.
            _include: The value of the include argument.
            _name: The value of the name argument.
            _parent: The value of the parent argument.
        """
        self._name = name
        self._data_source = config_source
        self._data_table = self._data_source.get_table(self._name)
        self._parent = parent
        self._include = include

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def __getattr__(self, attr: str) -> str:
        # First check this configuration to see if it has the requested attribute
        if values := self._data_table.get_rows(attr):
            return values[0].get_value(attr)

        # Next check the include configuration without considering its parents
        if self._include:
            include_parent_value = self._include._parent  # pylint: disable=protected-access
            self._include._parent = None  # pylint: disable=protected-access
            if hasattr(self._include, attr):
                self._include._parent = include_parent_value  # pylint: disable=protected-access
                return getattr(self._include, attr)
            self._include._parent = include_parent_value  # pylint: disable=protected-access

        # Next check the parent configuration considering its includes
        if self._parent and attr != ConfigCollection.INCLUDE_CONFIG_TAG:
            if hasattr(self._parent, attr):
                return getattr(self._parent, attr)

        # Finally check the include configuration considering its parents
        if self._include:
            if hasattr(self._include, attr):
                return getattr(self._include, attr)

        # The attribute wasn't found
        raise AttributeError(f'Unknown parameter ({attr}) for configuration: {self._name}')

    def __setattr__(self, attr: str, value: str) -> None:
        if attr.startswith('_'):
            super().__setattr__(attr, value)
            return
        self._data_table.get_rows()[0].setvalue(attr, value)
        self._data_source.commit()

    name = property(lambda s: s._name, doc='A read-only property which returns the name of the configuration.')



"""This module provides utilities for working with network services."""

# Import standard modules
from smtplib import SMTP
from typing import Dict, Optional, Tuple, Callable

# Import third-party modules
from requests import get as url_get, PreparedRequest
from requests.auth import AuthBase

# Import internal modules
from .lang import DEFAULT_ENCODING, is_debug


def download(url: str, /, target: Optional[str] = None, *, auth: Optional[Tuple[str, str] | AuthBase | Callable[[PreparedRequest], PreparedRequest]] = None) -> None:
    """Download a file from a URL target.

    Args:
        url: The URL to download from.
        target (optional, default=None): The file to which to download.
            If None, the last part of the URL will be used.
        auth (optional, default=None): If not None, it must be a (username, password) tuple.

    Returns:
        Nothing.

    Raises:
        Uses the requests module raise_for_status() function to raise on download errors.
    """
    target = target if target is not None else url.split('/')[-1]
    response = url_get(url, auth=auth, timeout=60)
    response.raise_for_status()
    with open(target, 'wb', encoding=DEFAULT_ENCODING) as downloaded_file:
        downloaded_file.write(response.content)


def send_email(smtp_server: str, receiver: str, sender: str, subject: str, body: str, /,
               content_type: str = 'text/plain') -> Dict[str, Tuple[int, bytes]]:
    """Send an SMTP email message.

    Args:
        smtp_server: The SMTP server to send the email through.
        receiver: The email address to which to send.
        sender: The return address for the email.
        subject: The email message subject.
        body: The email message body.
        content_type (optional, default='text/plain'): The content type for the email.

    Returns:
        The result of the sendmail call.
    """
    full_message = f'From: {sender}\nTo: {receiver}\nSubject: {subject}\nContent-Type: {content_type}\n\n' + '\n'.join(body)
    server = SMTP(smtp_server)
    if is_debug('SMTP'):
        server.set_debuglevel(True)
    receiver_list = receiver.split(',') if isinstance(receiver, str) else [receiver]
    try:
        result = server.sendmail(sender, receiver_list, full_message)
    finally:
        server.quit()
    return result



"""Unit tests for the sysutil module."""

# pylint: disable=missing-class-docstring,missing-function-docstring,invalid-name
# flake8: noqa

from enum import Enum
from multiprocessing import Process, Queue
from os import fdopen, getenv
from pathlib import Path
from stat import S_IRGRP, S_IROTH, S_IRUSR
from sys import platform
from tempfile import mkdtemp, mkstemp
from unittest import main, TestCase

from batcave.sysutil import (chmod, CMDError, get_app_config_dir, get_app_data_dir,
                             LockError, LockFile, LockMode, OSUtilError,
                             popd, pushd, rmpath, rmtree_hard, syscmd, SysCmdRunner)

LockSignal = Enum('LockSignal', ('true', 'false'))


class TestAppDirs(TestCase):
    APP_NAME = 'TestBatCaveApp'

    def test_data_dir_return_type(self):
        self.assertIsInstance(get_app_data_dir(self.APP_NAME), Path)

    def test_config_dir_return_type(self):
        self.assertIsInstance(get_app_config_dir(self.APP_NAME), Path)

    def test_data_dir_app_name(self):
        self.assertEqual(get_app_data_dir(self.APP_NAME).name, self.APP_NAME)

    def test_config_dir_app_name(self):
        self.assertEqual(get_app_config_dir(self.APP_NAME).name, self.APP_NAME)

    def test_data_dir_platform_path(self):
        match platform:
            case 'win32':
                expected = Path(getenv('LOCALAPPDATA', '')) / self.APP_NAME
            case 'darwin':
                expected = Path.home() / 'Library/Application Support' / self.APP_NAME
            case _:
                expected = Path(getenv('XDG_DATA_HOME', Path.home() / '.local/share')) / self.APP_NAME
        self.assertEqual(get_app_data_dir(self.APP_NAME), expected)

    def test_config_dir_platform_path(self):
        match platform:
            case 'win32':
                expected = Path(getenv('APPDATA', '')) / self.APP_NAME
            case 'darwin':
                expected = Path.home() / 'Library/Preferences' / self.APP_NAME
            case _:
                expected = Path(getenv('XDG_CONFIG_HOME', Path.home() / '.config')) / self.APP_NAME
        self.assertEqual(get_app_config_dir(self.APP_NAME), expected)


class TestChmod(TestCase):
    def test_chmod_directory(self):
        tempdir = Path(mkdtemp())
        try:
            chmod(tempdir, S_IRUSR | S_IRGRP | S_IROTH)
            mode = tempdir.stat().st_mode
            match platform:
                case 'win32':
                    pass  # Windows chmod is limited
                case _:
                    self.assertTrue(mode & S_IRUSR)
        finally:
            tempdir.chmod(0o700)
            tempdir.rmdir()

    def test_chmod_recursive(self):
        tempdir = Path(mkdtemp())
        subdir = tempdir / 'sub'
        subdir.mkdir()
        testfile = subdir / 'file.txt'
        testfile.write_text('test')
        try:
            chmod(tempdir, 0o755, recursive=True)
            match platform:
                case 'win32':
                    pass  # Windows chmod is limited
                case _:
                    self.assertTrue(subdir.stat().st_mode & S_IRUSR)
                    self.assertTrue(testfile.stat().st_mode & S_IRUSR)
        finally:
            rmtree_hard(tempdir)

    def test_chmod_files_only(self):
        tempdir = Path(mkdtemp())
        testfile = tempdir / 'file.txt'
        testfile.write_text('test')
        try:
            original_dir_mode = tempdir.stat().st_mode
            chmod(tempdir, 0o644, recursive=True, files_only=True)
            self.assertEqual(tempdir.stat().st_mode, original_dir_mode)
        finally:
            rmtree_hard(tempdir)


class TestDirStack(TestCase):
    def setUp(self):
        self._tempdir = Path(mkdtemp()).resolve()

    def tearDown(self):
        self._tempdir.rmdir()

    def test_popd_empty_stack(self):
        self.assertEqual(popd(), 0)

    def test_push_and_pop(self):
        start = Path.cwd()

        old_dir = pushd(self._tempdir)
        new_dir = Path.cwd()
        self.assertEqual(old_dir, start)
        self.assertEqual(new_dir, self._tempdir)

        old_dir = popd()
        new_dir = Path.cwd()
        self.assertEqual(old_dir, start)
        self.assertEqual(new_dir, start)


class TestExceptions(TestCase):
    def test_PlatformException(self):
        try:
            raise CMDError(CMDError.INVALID_OPERATION, platform='bad-platform')
        except CMDError as err:
            self.assertEqual(CMDError.INVALID_OPERATION.code, err.code)


class TestLockFile(TestCase):
    def setUp(self):
        (fd, fn) = mkstemp()
        self._fh = fdopen(fd)
        self._fn = Path(fn)
        self._got_lock = Queue()
        self._lock_again = Process(target=secondary_lock_process, args=(self._fn, self._got_lock))

    def tearDown(self):
        self._got_lock.close()
        self._fh.close()
        if self._fn.exists():
            self._fn.unlink()

    def test_1_cleanup(self):
        with LockFile(self._fn, handle=self._fh, cleanup=True):
            pass
        self.assertFalse(self._fn.exists())

    def test_2_no_cleanup(self):
        with LockFile(self._fn, handle=self._fh, cleanup=False):
            pass
        self.assertTrue(self._fn.exists())

    def test_3_lock(self):
        with LockFile(self._fn, handle=self._fh, cleanup=True):
            self._lock_again.start()
            got_lock = self._got_lock.get()
            self._lock_again.join()
            self.assertTrue(got_lock == LockSignal.false)

    def test_4_unlock(self):
        with LockFile(self._fn, handle=self._fh, cleanup=True) as lockfile:
            lockfile.action(LockMode.unlock)
            self._lock_again.start()
            got_lock = self._got_lock.get()
            self._lock_again.join()
            lockfile.action(LockMode.lock)
            self.assertTrue(got_lock == LockSignal.true)


def secondary_lock_process(filename, queue):
    try:
        with LockFile(filename, cleanup=False) as lock_again:
            lock_again.action(LockMode.unlock)
            queue.put(LockSignal.true)
    except LockError:
        queue.put(LockSignal.false)


class TestOSUtilError(TestCase):
    def test_group_exists(self):
        try:
            raise OSUtilError(OSUtilError.GROUP_EXISTS, group='test_group')
        except OSUtilError as err:
            self.assertEqual(OSUtilError.GROUP_EXISTS.code, err.code)

    def test_invalid_operation(self):
        try:
            raise OSUtilError(OSUtilError.INVALID_OPERATION, platform='bad-platform')
        except OSUtilError as err:
            self.assertEqual(OSUtilError.INVALID_OPERATION.code, err.code)

    def test_user_exists(self):
        try:
            raise OSUtilError(OSUtilError.USER_EXISTS, user='testuser')
        except OSUtilError as err:
            self.assertEqual(OSUtilError.USER_EXISTS.code, err.code)


class TestRmPath(TestCase):
    def test_remove_file(self):
        (fd, fn) = mkstemp()
        fdopen(fd).close()
        path = Path(fn)
        self.assertTrue(path.exists())
        rmpath(path)
        self.assertFalse(path.exists())

    def test_remove_directory(self):
        tempdir = Path(mkdtemp())
        self.assertTrue(tempdir.exists())
        rmpath(tempdir)
        self.assertFalse(tempdir.exists())

    def test_remove_directory_with_contents(self):
        tempdir = Path(mkdtemp())
        (tempdir / 'subdir').mkdir()
        (tempdir / 'subdir' / 'file.txt').write_text('test')
        (tempdir / 'file.txt').write_text('test')
        rmpath(tempdir)
        self.assertFalse(tempdir.exists())


class TestRmtreeHard(TestCase):
    def test_remove_tree(self):
        tempdir = Path(mkdtemp())
        (tempdir / 'subdir').mkdir()
        (tempdir / 'subdir' / 'file.txt').write_text('test')
        rmtree_hard(tempdir)
        self.assertFalse(tempdir.exists())

    def test_remove_readonly_tree(self):
        tempdir = Path(mkdtemp())
        readonly_file = tempdir / 'readonly.txt'
        readonly_file.write_text('test')
        readonly_file.chmod(S_IRUSR | S_IRGRP | S_IROTH)
        rmtree_hard(tempdir)
        self.assertFalse(tempdir.exists())


class TestSysCmd(TestCase):
    def test_basic_command(self):
        match platform:
            case 'win32':
                result = syscmd('cmd', '/c', 'echo', 'hello')
            case _:
                result = syscmd('echo', 'hello')
        self.assertTrue(any('hello' in line for line in result))

    def test_flatten_output(self):
        match platform:
            case 'win32':
                result = syscmd('cmd', '/c', 'echo', 'hello', flatten_output=True)
            case _:
                result = syscmd('echo', 'hello', flatten_output=True)
        self.assertIsInstance(result, str)
        self.assertIn('hello', result)

    def test_cmd_not_found(self):
        with self.assertRaises((CMDError, FileNotFoundError)):
            syscmd('this_command_does_not_exist_batcave')

    def test_remote_options_without_remote(self):
        with self.assertRaises(CMDError) as context:
            syscmd('echo', 'hello', remote_auth=('user', 'pass'))
        self.assertEqual(CMDError.INVALID_OPERATION.code, context.exception.code)


class TestSysCmdRunner(TestCase):
    def test_basic_run(self):
        match platform:
            case 'win32':
                runner = SysCmdRunner('cmd', '/c', 'echo', show_cmd=False, show_stdout=False)
                result = runner.run('hello')
            case _:
                runner = SysCmdRunner('echo', show_cmd=False, show_stdout=False)
                result = runner.run('hello')
        self.assertTrue(any('hello' in line for line in result))


if __name__ == '__main__':
    main()

# cSpell:ignore batcave localappdata syscmd



"""This module provides a simplified interface to the standard platform module."""
# pylint: disable-all

# Import standard modules
from enum import Enum
from pathlib import Path
from platform import uname
from sys import version_info

OsType = Enum('OsType', ('linux', 'windows'))


class Platform:
    """A class to provide a simplified interface to the platform and sys.version_info standard modules."""

    def __getattr__(self, attr: str) -> str:
        """Get the platform type formatted for the requested subtype."""
        sys_info = uname()
        batcave_os = bart_os = sys_info.system.replace('-', '')
        batcave_version = batcave_arch = ''
        bart_version = bart_arch = ''
        p4ver = p4_arch = du_os = build = ''
        match batcave_os:
            case 'AIX':
                batcave_version = bart_version = sys_info.version + sys_info.release
                batcave_arch = bart_arch = 'ppc'
                build = f'{sys_info.system} {sys_info.version}.{sys_info.release} PowerPC'
                p4ver = batcave_version
            case 'HPUX':
                os_major = sys_info.release.split('.')[1]
                os_minor = sys_info.release.split('.')[2]
                batcave_version = bart_version = os_major + os_minor
                batcave_arch = bart_arch = sys_info.machine.split('/')[0]
                build = f'{sys_info.system} {sys_info.release} {sys_info.machine}'
                p4ver = os_major
            case 'Linux':
                batcave_version = bart_version = sys_info.release.split('.')[0] + sys_info.release.split('.')[1]
                bart_arch = sys_info.machine
                if sys_info.machine == 'x86_64':
                    batcave_arch = 'i686'
                else:
                    batcave_arch = sys_info.machine
                try:
                    build = open([f for f in Path('/etc').glob('*-release')][0]).readline().strip()
                except IndexError:
                    build = 'unknown'

                p4ver = batcave_version
                if (p4_arch := sys_info.processor.replace(' ', '_')) in ('i686', 'i386', 'athalon'):
                    p4_arch = 'x86'
            case 'Darwin':
                batcave_version = bart_version = sys_info.release.split('.')[0] + sys_info.release.split('.')[1]
                batcave_arch = bart_arch = sys_info.machine
                build = f'{sys_info.system} {sys_info.release} {sys_info.machine}'
                p4ver = '80'
                if (p4_arch := sys_info.processor.replace(' ', '_')) in ('i686', 'i386', 'athalon'):
                    p4_arch = 'x86'
            case 'Windows':
                (batcave_version, batcave_arch, p4ver, bart_version, bart_arch) = ('',) * 5
                batcave_os = 'win32'
                du_os = 'win'
                if sys_info.machine.endswith('64'):
                    p4_arch = 'x64'
                    bart_os = 'win64'
                else:
                    p4_arch = 'x86'
                    bart_os = 'win32'
                build = f'{sys_info.system} {sys_info.release} {p4_arch[1:]}-bit ({sys_info.version})'

        match attr:
            case 'bart':
                return bart_os + bart_version + bart_arch
            case 'distutils':
                return '%s-%s-%s' % (du_os, uname().machine.lower(), '.'.join([str(i) for i in version_info[:2]]))
            case 'batcave_run':
                return batcave_os + batcave_version + batcave_arch
            case 'batcave_build':
                return build
            case 'p4':
                return '%s%s%s' % (batcave_os.lower().replace('windows', 'nt'), p4ver, p4_arch)
        raise AttributeError(f'Unknown platform type: {attr}')

# cSpell:ignore batcave



"""This module provides utilities for managing cloud resources.

Attributes:
    CloudType (Enum): The cloud providers supported by the Cloud class.
    gcloud (SysCmdRunner.run): A simple interface to the gcloud command line tool.
"""

# Import standard modules
from ast import literal_eval
from enum import Enum
from os import getenv
from pathlib import Path
from string import Template
from typing import Any, List, Optional, Sequence

# Import third-party modules
from docker import DockerClient
from docker.errors import ImageNotFound
from docker.models.containers import Container as DockerContainer
from requests import head as get_head

# Import internal modules
from .lang import BatCaveError, BatCaveException, CommandResult, WIN32
from .sysutil import SysCmdRunner

CloudType = Enum('CloudType', ('local', 'gcloud', 'dockerhub'))

# pylint: disable=invalid-name
if WIN32:
    user_install = Path(getenv('USERPROFILE', '')) / 'APPDATA/LOCAL'
    gcloud_command_location = 'Google/Cloud SDK/google-cloud-sdk/bin/gcloud.cmd'
    if (user_install / gcloud_command_location).exists():
        gcloud_command = str(user_install / gcloud_command_location)
    else:
        gcloud_command = str(Path(getenv('ProgramFiles(x86)', '')) / gcloud_command_location)
else:
    gcloud_command = 'gcloud'
gcloud = SysCmdRunner(gcloud_command, quiet=True, syscmd_args={'use_shell': WIN32}).run
# pylint: enable=invalid-name


class CloudError(BatCaveException):
    """Cloud Exceptions.

    Attributes:
        IMAGE_ERROR: There was an error working with a container image.
        INVALID_OPERATION: The specified cloud type does not support the requested operation.
        INVALID_TYPE: An invalid cloud type was specified.
    """
    IMAGE_ERROR = BatCaveError(1, Template('Error ${action}ing image: $err'))
    INVALID_OPERATION = BatCaveError(2, Template('Invalid Cloud type ($ctype) for this operation'))
    INVALID_TYPE = BatCaveError(3, Template('Invalid Cloud type ($ctype). Must be one of: ' + str([t.name for t in CloudType])))


class Cloud:
    """Class to create a universal abstract interface for a cloud instance."""

    def __init__(self, ctype: CloudType, /, *, auth: Sequence[str] = ('', ''), login: bool = True):
        """
        Args:
            ctype: The cloud provider for this instance. Must be a member of CloudType.
            auth (optional, default=None): This is a (username, password) tuple.
            login (optional, default=True): Whether or not to login to the cloud provider at instance initialization.

        Attributes:
            auth: The value of the auth argument.
            _client: A reference to the underlying client API object.
            _type: The value of the ctype argument.
        """
        self._client: Any = False
        self._type = ctype
        self.auth = (auth[0], auth[1])
        validate_type(self.type)
        if login:
            self.login()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    client = property(lambda s: s._client, doc='A read-only property which returns the client reference.')
    type = property(lambda s: s._type, doc='A read-only property which returns the cloud type.')

    def exec(self, *args, **kwargs) -> CommandResult:
        """Execute a command against the cloud API.

        Args:
            *args (optional, default=[]): A list of arguments to pass to the API.
            **kwargs (optional, default={}): A dictionary to pass to the API.

        Returns:
            The result of the API call.

        Raises:
            CloudError.INVALID_OPERATION: If the cloud type does not support an API call.
        """
        match self.type:
            case CloudType.gcloud:
                return gcloud(*args, **kwargs)
        raise CloudError(CloudError.INVALID_OPERATION, ctype=self.type.name)

    def get_container(self, name: str, /) -> 'Container':
        """Get a container from the cloud.

        Args:
            name: The container name to retrieve.

        Returns:
            The container object.
        """
        return Container(self, name)

    def get_containers(self, filters: Optional[str] = None, /) -> List['Container']:
        """Get a possibly filtered list of containers.

        Args:
            filter (optional, default=None): the container name to retrieve.

        Returns:
            The container object.

        Raises:
            CloudError.INVALID_OPERATION: If the cloud type does not support login.
        """
        match self.type:
            case CloudType.local | CloudType.dockerhub:
                return [Container(self, c.name) for c in self._client.containers.list(filters=filters)]
        raise CloudError(CloudError.INVALID_OPERATION, ctype=self.type.name)

    containers = property(get_containers, doc='A read-only property which calls the get_containers() method with no filters.')

    def get_image(self, tag: str, /) -> 'Image':
        """Get an image from the cloud container registry.

        Args:
            tag: The container image tag to retrieve.

        Returns:
            The image object.
        """
        return Image(self, tag)

    def login(self) -> None:
        """Perform a login to the cloud provider.

        Returns:
            Nothing.
        """
        self._client = DockerClient()
        if self.type == CloudType.dockerhub:
            self._client.login(*self.auth)


class Image:
    """Class to create a universal abstract interface to a container image."""

    def __init__(self, cloud: Cloud, name: str, /):
        """
        Args:
            cloud: The API cloud reference.
            name: The image name.

        Attributes:
            _cloud: The value of the cloud argument.
            _docker_client: A reference to the client from the Docker API.
            _name: The value of the name argument.
            _ref: A reference to the underlying API object.
        """
        self._cloud = cloud
        self._name = name
        self._docker_client: DockerClient = self.cloud.client if isinstance(self.cloud.client, DockerClient) else DockerClient()
        if self._cloud.type == CloudType.dockerhub:
            (image, tag) = self.name.split(':', 1)
            response = get_head(f'https://registry.hub.docker.com/v2/{image}/manifests/{tag}', auth=self._cloud.auth,
                                headers={"Accept": "application/vnd.docker.distribution.manifest.v2+json"}, timeout=10)
            if response.status_code != 200:
                self._ref = None
        try:
            self._ref = self.cloud.client.images.get(self.name)
        except ImageNotFound:
            self._ref = None

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    cloud = property(lambda s: s._cloud, doc='A read-only property which returns the container cloud object.')
    name = property(lambda s: s._name, doc='A read-only property which returns the name of the image.')

    containers = property(lambda s: s.cloud.get_containers({'ancestor': s.name}),
                          doc='A read-only property which returns all the containers for this image.')

    def get_tags(self) -> List[str]:
        """Get a list of tags applied to the image.

        Returns:
            The sorted list of tags applied to the image.
        """
        return self._ref.tags if self._ref else []

    tags = property(get_tags, doc='A read-only property which calls the get_tags() method with no filters.')

    def pull(self) -> 'Image':
        """Pull the image from the registry.

        Returns:
            The image object.
        """
        self._ref = self._docker_client.images.pull(self.name)
        return self

    def push(self) -> List[str]:
        """Push the image to the registry.

        Returns:
            The server log from the push.
        """
        docker_log = [literal_eval(line.strip()) for line in self._docker_client.images.push(self.name).splitlines() if line]
        errors = [line['error'] for line in docker_log if 'error' in line]
        if errors:
            raise CloudError(CloudError.IMAGE_ERROR, action='push', err=''.join(errors))
        return docker_log

    def run(self, *, detach: bool = True, update: bool = True, **kwargs) -> DockerContainer:
        """Run an image to create an active container.

        Args:
            detach (optional, default=True): If True, do not wait for the container to complete.
            update (optional, default=True): If True, perform a pull of the image from the registry before running.
            **kwargs (optional, default={}): A dictionary of arguments to pass to the run command.

        Returns:
            A reference to the active container.

        Raises:
            CloudError.INVALID_OPERATION: If the cloud type does not support running an image.
        """
        if update:
            self.pull()
        match self.cloud.type:
            case CloudType.local | CloudType.dockerhub:
                return self.cloud.client.containers.run(self.name, detach=detach, **kwargs)
        raise CloudError(CloudError.INVALID_OPERATION, ctype=self.cloud.type.name)

    def tag(self, new_tag: str, /) -> Optional['Image']:
        """Tag an image in the registry.

        Returns:
            The tagged image.
        """
        new_ref: Optional[Image] = None
        self.pull()
        if self._ref is None:
            raise CloudError(CloudError.IMAGE_ERROR, action='tag', err='Image not found')
        self._ref.tag(new_tag)
        (new_ref := Image(self.cloud, new_tag)).push()
        return new_ref


class Container:
    """Class to create a universal abstract interface to a container."""

    def __init__(self, cloud: Cloud, name: str, /):
        """
        Args:
            cloud: The API cloud reference.
            name: The container name.

        Attributes:
            _cloud: The value of the cloud argument.
            _name: The value of the name argument.
            _ref: A reference to the underlying API object.

        Raises:
            CloudError.INVALID_OPERATION: If the specified cloud type is not supported.
        """
        self._cloud = cloud
        self._name = name
        self._ref: Any = None
        match self.cloud.type:
            case CloudType.local | CloudType.dockerhub:
                self._ref = self.cloud.client.containers.get(self.name)
            case _:
                raise CloudError(CloudError.INVALID_OPERATION, ctype=self.cloud.type.name)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    cloud = property(lambda s: s._cloud, doc='A read-only property which returns the container cloud object.')
    name = property(lambda s: s._name, doc='A read-only property which returns the name of the container.')

    def stop(self) -> DockerContainer:
        """Stop a running container.

        Returns:
            A reference to the stopped container.

        Raises:
            CloudError.INVALID_OPERATION: If the cloud type does not support stopping an container.
        """
        match self.cloud.type:
            case CloudType.local | CloudType.dockerhub:
                return self._ref.stop()
        raise CloudError(CloudError.INVALID_OPERATION, ctype=self.cloud.type.name)


def validate_type(ctype: CloudType) -> None:
    """Determine if the specified Cloud type is valid.

    Args:
        ctype: The Cloud type.

    Returns:
        Nothing.

    Raises
        CloudError.INVALID_TYPE: If the cloud type is not valid.
    """
    if ctype not in CloudType:
        raise CloudError(CloudError.INVALID_TYPE, ctype=ctype)

# cSpell:ignore dockerhub syscmd



"""This module provides utilities for managing data sources.

Data source definitions:
    TEXT
        source - a text file with '>NAME' delimiting tables
        table - a list of rows
        row - a line of the format column1:value1|column2:value2
        column - delimited part of the line

    INI
        source - an ini file

        table - an ini section of the form::
            [TABLENAME]
            ROWS: 1,...,N

        row - an ini section of the form::
            [TABLENAME ROW N]
            col1: val1
            col2: val2

        column - a section option

    PICKLE
        source - a Python pickle file containing a single dictionary
        table - a member of the top level dictionary which is a list of rows
        row - a dictionary of column:value pairs

    XML_SINGLE
        This format is a kludge to support the old RCFile format
        source - an XML file with the top level element the name of the datasource
        table - there is only one table which has the same name as the top level element

        row - an XML element of the form::
            <environment name=COL-NAME-VALUE>
            COLUMNS
            </environment>

        column - an XML element of the form::
            <COL-NAME>COL-VALUE</COL-NAME>

    XML_FLAT
        source - an XML file with the top level element the name of the datasource
        table - a collection of XML elements

        row - an XML element of the form::
            <TABLE-NAME COL-NAME="COL-VAL" />

        column - an XML element attribute

    XML
        source - an XML file with the top level element the name of the datasource

        table - an XML element of the form::
            <TABLE name="TABLE-NAME">
            ROWS
            </TABLE>

        row - an XML element of the form::
            <ROW>
            COLUMNS
            </ROW>

        column - an XML element of the form::
            <COL-NAME>COL-VALUE</COL-NAME>

Attributes:
    SourceType (Enum): The source providers supported by the DataSource class.
"""

# Import standard modules
from configparser import RawConfigParser
from enum import Enum
from pathlib import PurePath
from pickle import dump as pickle_dump, load as pickle_load
from string import Template
from typing import cast, Any, IO, List, Optional
from urllib.request import urlopen
from xml.parsers import expat
from xml.etree.ElementTree import ElementTree
import xml.etree.ElementTree as xml_etree

# Import internal modules
from .lang import DEFAULT_ENCODING, BatCaveError, BatCaveException, PathName

SourceType = Enum('SourceType', ('text', 'ini', 'pickle', 'xml'))


class DataError(BatCaveException):
    """Data related Exceptions.

    Attributes:
        BAD_COLUMN: The column specified was not found.
        BAD_ROOT: The root element for the XML data source was not the one requested.
        BAD_SCHEMA: The data source schema does not match the one requested.
        BAD_TABLE: The requested table was not found.
        BAD_URL: The URL specified does not contain a valid data source.
        FILE_OPEN: Error opening the data source file.
        INVALID_OPERATION: The specified data source type does not support the requested operation.
        INVALID_TYPE: An invalid data source type was specified.
    """
    BAD_COLUMN = BatCaveError(1, Template('Invalid column: $col\nonline $line'))
    BAD_ROOT = BatCaveError(2, Template('The root element "$root_name" is not the one requested ($expected) in $source_name'))
    BAD_SCHEMA = BatCaveError(3, Template('Wrong schema ($schema) specified for data source ($found)'))
    BAD_TABLE = BatCaveError(4, Template('No table named "$table_name" in data source "$source_name"'))
    BAD_URL = BatCaveError(5, Template('No valid DataSource found at URL: $url'))
    FILE_OPEN = BatCaveError(6, Template('Unable to open file: $err_msg'))
    INVALID_OPERATION = BatCaveError(7, Template('Function "$function" not supported for source type "$source_type"'))
    INVALID_TYPE = BatCaveError(8, 'Invalid data source type. Must be one of: ' + str([t.name for t in SourceType]))


class DataSource:
    """Class to create a universal abstract interface for a data source.

    Attributes:
        INFO_TABLE: The table in the data source which has info about the source.
        _SCHEMA_DATA: The row which contains the schema version.

        INI_ROW_LIST_OPT: The INI file section which contains the rows.
        INI_ROW_TAG: The INI file tag to mark a row.

        _TEXT_TABLE_DELIMITER: The table delimiter for a text file.
        _TEXT_VAL_DELIMITER: The value delimiter for a text file.

        _XML_TABLE_TAG: The XML file tag to denote a table.
        _XML_TABLE_NAME_ATTRIBUTE: The XML file attribute to denote a table name.
        _XML_SINGLE_COL_NAME: The XML file attribute to denote a column name.
    """
    _SCHEMA_DATA = 'schema'
    _TEXT_TABLE_DELIMITER = '>'
    _TEXT_VAL_DELIMITER = ':'
    _XML_TABLE_TAG = 'TABLE'
    _XML_TABLE_NAME_ATTRIBUTE = 'name'
    _XML_SINGLE_COL_NAME = 'name'

    INFO_TABLE = 'DataSourceInfo'
    INI_ROW_TAG = ' ROW '
    INI_ROW_LIST_OPT = 'ROWS'

    def __init__(self, data_type: SourceType, connect_info: PathName, /, name: str, *, schema: int, create: bool = False):
        """
        Args:
            data_type: The data source type.
            connect_info: The data source file name.
            name: The data source name.
            schema: The schema version of the data source.
            create (optional, default=False): If True, the data source will be created if it does not exist.

        Attributes:
            _closer: This is the method used to close the data source.
            _connect_info: The value of the connect_info argument.
            _connection: For XML_* data source types, this it the parsed XML tree,
                otherwise it is a temporary file object.
            _name: The value of the name argument.
            _schema: The value of the schema argument.
            _source: The value depends on the data source type.
                TEXT: the file contents
                PICKLE: the top dictionary
                INI: The RawConfigParser object from the INI file
                XML_*: The XML root of the file
            _type: The value of the data_type argument.

        Raises:
            DataError.FILE_OPEN: If the data source file is already open.
        """
        self._name = name
        self._schema = schema
        self._source: Any = None
        self._type = data_type
        self._connect_info = connect_info
        self._connection: Optional[IO | ElementTree] = None
        self._closer: Optional[IO] = None
        self._validate_type()
        try:
            self._load()
            self._validate_schema()
        except IOError as ioe:
            if create and (ioe.errno == 2):
                self._create()
            else:
                raise DataError(DataError.FILE_OPEN, err_msg=str(ioe)) from ioe

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()
        return False

    def _create(self) -> None:
        """Create the data source.

        Returns:
            Nothing.
        """
        match self.type:
            case SourceType.text | SourceType.pickle:
                self._source = {}
            case SourceType.ini:
                self._source = RawConfigParser()
            case SourceType.xml:
                self._source = xml_etree.Element(self.name)
                self._connection = xml_etree.ElementTree(self._source)
        self.add_table(self.INFO_TABLE).add_row(schema=str(self._schema))
        self.commit()

    def _load(self) -> None:  # pylint: disable=too-many-branches
        """Load the data source.

        Returns:
            Nothing.

        Raises:
            DataError.BAD_COLUMN: If a text type data source has a column with no value.
            DataError.BAD_ROOT: If the root of the data source does not match the name.
            DataError.BAD_URL: If the data source is specified as a malformed URL.
        """
        match self.type:
            case SourceType.text:
                self._source = {}
                table_name: str = ''
                with open(self._connect_info, encoding=DEFAULT_ENCODING) as data_source:
                    for raw_line in data_source:
                        if (line := raw_line.strip()).startswith(self._TEXT_TABLE_DELIMITER):
                            self._source[(table_name := line.lstrip(self._TEXT_TABLE_DELIMITER))] = []
                            continue
                        row = {}
                        for pair in line.split(self._TEXT_TABLE_DELIMITER):
                            try:
                                (col, val) = pair.split(self._TEXT_VAL_DELIMITER)
                            except ValueError as err:
                                raise DataError(DataError.BAD_COLUMN, col=pair, line=line) from err
                            row[col] = val
                        self._source[table_name].append(row)
            case SourceType.pickle:
                with open(self._connect_info, encoding=DEFAULT_ENCODING) as data_source:
                    self._source = pickle_load(cast(IO[bytes], data_source))
            case SourceType.ini:
                self._source = RawConfigParser()
                with open(self._connect_info, encoding=DEFAULT_ENCODING) as ini_tmp:
                    self._source.read_file(ini_tmp)
            case SourceType.xml:
                if isinstance(self._connect_info, list):
                    self._connection = xml_etree.ElementTree(xml_etree.fromstring(' '.join(self._connect_info)))
                else:
                    if isinstance(self._connect_info, (str, PurePath)):
                        if str(self._connect_info).startswith('http:') or str(self._connect_info).startswith('file:'):
                            self._closer = urlopen(cast(str, self._connect_info))  # pylint: disable=consider-using-with
                        else:
                            self._closer = open(self._connect_info, encoding=DEFAULT_ENCODING)  # pylint: disable=consider-using-with
                    else:
                        self._closer = self._connect_info
                    try:
                        self._connection = cast(ElementTree, xml_etree.parse(cast(bytes, self._closer)))
                    except expat.ExpatError as err:
                        if expat.ErrorString(err.code) == expat.errors.XML_ERROR_SYNTAX:  # pylint: disable=no-member
                            raise DataError(DataError.BAD_URL, url=self._connect_info) from err
                        raise
                self._source = cast(ElementTree, self._connection).getroot()
                if self._source.tag != self.name:
                    raise DataError(DataError.BAD_ROOT, source_name=self._connect_info, root_name=self._source.tag, expected=self.name)

    def _validate_schema(self) -> None:
        """Determine if the data source schema is valid.

        Returns:
            Nothing.

        Raises:
            DataError.BAD_SCHEMA: If the data source schema is invalid.
        """
        if self._schema != self.schema:
            raise DataError(DataError.BAD_SCHEMA, schema=self._schema, found=self.schema)

    def _validate_type(self) -> None:
        """Determine if the specified data source type is valid.

        Returns:
            Nothing.

        Raises:
            DataError.INVALID_TYPE: If the data source type is invalid.
        """
        if self.type not in SourceType:
            raise DataError(DataError.INVALID_TYPE)

    filename = property(lambda s: s._connect_info, doc='A read-only property which returns the file name of the data source.')
    name = property(lambda s: s._name, doc='A read-only property which returns the name of the data source.')
    type = property(lambda s: s._type, doc='A read-only property which returns the type of the data source.')

    @property
    def dict_repr(self) -> dict:
        """A read-only property which returns a the data source as a dictionary."""
        dict_repr = {}
        for table in self.get_tables():
            with table.get_rows()[0] as row:
                dict_repr[table.name] = {c: row.get_value(c) for c in row.get_columns()}
        return dict_repr

    @property
    def schema(self) -> int:
        """A read-only property which returns the schema value of the data source."""
        try:
            return int(self.get_table(self.INFO_TABLE).get_rows(col=self._SCHEMA_DATA)[0].get_value(self._SCHEMA_DATA))
        except IndexError:
            return 0
        except DataError as err:
            if err.code != DataError.BAD_TABLE.code:
                raise
            return 0

    def add_table(self, name: str, /) -> 'DataTable':
        """Create a new table with the specified name.

        Args:
            name: The name of the table to add.

        Returns:
            The created table.

        Raises:
            DataError.INVALID_OPERATION: If the data source type does not support adding a table.
        """
        match self.type:
            case SourceType.text | SourceType.pickle:
                self._source[name] = []
            case SourceType.ini:
                self._source.add_section(name)
                self._source.set(name, self.INI_ROW_LIST_OPT, '')
            case SourceType.xml:
                table = xml_etree.SubElement(self._source, self._XML_TABLE_TAG)
                table.attrib[self._XML_TABLE_NAME_ATTRIBUTE] = name
        return self.get_table(name)

    def close(self) -> None:
        """Close the data source.

        Returns:
            Nothing.
        """
        if self._closer:
            self._closer.close()
            self._connect_info = ''
            self._closer = self._source = self._connection = None

    def commit(self) -> None:
        """Commit any changes in memory to the source.

        Returns:
            Nothing.
        """
        match self.type:
            case SourceType.text:
                self._connection = open(self._connect_info, 'w', encoding=DEFAULT_ENCODING)  # pylint: disable=consider-using-with
                for (table_name, rows) in self._source.items():
                    self._connection.write(self._TEXT_TABLE_DELIMITER + table_name + '\n')
                    for row in rows:
                        self._connection.write(self._TEXT_TABLE_DELIMITER.join([f'{col}{self._TEXT_VAL_DELIMITER}{row[col]}' for col in row]) + '\n')
                self._connection.close()
            case SourceType.pickle:
                self._connection = open(self._connect_info, 'w', encoding=DEFAULT_ENCODING)  # pylint: disable=consider-using-with
                pickle_dump(self._source, cast(IO, self._connection))
                self._connection.close()
            case SourceType.ini:
                self._connection = open(self._connect_info, 'w', encoding=DEFAULT_ENCODING)  # pylint: disable=consider-using-with
                self._source.write(self._connection)
                self._connection.close()
            case SourceType.xml:
                cast(ElementTree, self._connection).write(cast(str, self._connect_info), 'ISO-8859-1')

    def get_table(self, name: str, /) -> 'DataTable':  # pylint: disable=too-many-branches
        """Get the requested data table.

        Args:
            name: The name of the table to return.

        Returns:
            The requested data table.

        Raises:
            DataError.BAD_TABLE: If the requested table is not found.
        """
        table = None
        match self.type:
            case SourceType.text | SourceType.pickle:
                if name in self._source:
                    table = self._source[name]
            case SourceType.ini:
                if self._source.has_section(name):
                    row_list = self._source.get(name, self.INI_ROW_LIST_OPT)
                    table = [int(r) for r in row_list.split(',')] if row_list else []
            case SourceType.xml:
                for tmp_table in self._source.iter(self._XML_TABLE_TAG):
                    if tmp_table.get(self._XML_TABLE_NAME_ATTRIBUTE) == name:
                        table = tmp_table
                        break
        if table is None:
            raise DataError(DataError.BAD_TABLE, table_name=name, source_name=self.name)
        return DataTable(self.type, name, table, self._source)

    def get_tables(self) -> List['DataTable']:
        """Get all the tables from the data source.

        Returns:
            The list of table in the data source.
        """
        table_names: List[str] = []
        match self.type:
            case SourceType.text | SourceType.pickle:
                table_names = [t for t in self._source]  # pylint: disable=unnecessary-comprehension
            case SourceType.ini:
                table_names = [t for t in self._source.sections() if self.INI_ROW_TAG not in t]
            case SourceType.xml:
                table_names = [t.get(self._XML_TABLE_NAME_ATTRIBUTE) for t in self._source.iter(self._XML_TABLE_TAG)]
        return [self.get_table(t) for t in table_names]

    def has_table(self, name: str, /) -> bool:
        """Determine if the named table exists in the data source.

        Args:
            name: The name of the table to search for.

        Returns:
            Returns True if the named table exists in the data source, False otherwise.
        """
        match self.type:
            case SourceType.text | SourceType.pickle:
                if name in self._source:
                    return True
            case SourceType.ini:
                if self._source.has_section(name):
                    return True
            case SourceType.xml:
                for tmp_table in self._source.iter(self._XML_TABLE_TAG):
                    if tmp_table.get(self._XML_TABLE_NAME_ATTRIBUTE) == name:
                        return True
        return False


class DataRow:
    """Class to create a universal abstract interface for a data row."""

    def __init__(self, data_type: SourceType, raw: Any, parent: Any, /):
        """
        Args:
            data_type: The data source type.
            raw: The raw value of the row as read from the data source.
            parent: This is a reference to the table containing the row.

        Attributes:
            _parent: The value of the parent argument.
            _row: The value of the raw argument.
            _type: The value of the data_type argument.
        """
        self._row = raw
        self._parent = parent
        self._type = data_type

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def __getattr__(self, attr: str) -> str:
        return self.get_value(attr)

    raw = property(lambda s: s._row, doc='A read-only property which returns the raw contents of the data row.')
    type = property(lambda s: s._type, doc='A read-only property which returns the type of the data source.')

    def del_column(self, col: str, ) -> None:
        """Delete the named column from the row.

        Args:
            col: The name of the column to delete.

        Returns:
            Nothing.
        """
        if not self.has_col(col):
            return

        match self.type:
            case SourceType.text | SourceType.pickle:
                del self._row[col]
            case SourceType.ini:
                self._parent.remove_option(col)
            case SourceType.xml:
                self._row.remove(self._row.find(col))

    def delete(self) -> None:
        """Delete this data row.

        Returns:
            Nothing.
        """
        match self.type:
            case SourceType.ini:
                self._parent.remove_section(self._row)
            case _:
                self._parent.remove(self._row)

    def get_columns(self) -> List[str]:
        """Get all the columns from the row.

        Returns:
            The list of columns in the data row.
        """
        match self.type:
            case SourceType.text | SourceType.pickle:
                return list(self._row.keys())
            case SourceType.ini:
                return self._parent.options(self._row)
            case SourceType.xml:
                return [e.tag for e in list(self._row) if e.tag != self._XML_ROW_TAG]
        return []

    def get_value(self, col: str, /) -> str:
        """Get the value of the specified column.

        Args:
            col: The name of the column from which to retrieve the value.

        Returns:
            The value of the column.
        """
        match self.type:
            case SourceType.text | SourceType.pickle:
                if self.has_col(col):
                    return self._row[col]
            case SourceType.ini:
                if self.has_col(col):
                    return self._parent.get(self._row, col)
            case SourceType.xml:
                return self._row.findtext(col)
        return ''

    def has_col(self, col: str, /) -> bool:  # pylint: disable=too-many-return-statements
        """Determine if the named column exists in the row.

        Args:
            col: The name of the column to search for.

        Returns:
            Returns True if the named column exists in the data source, False otherwise.
        """
        match self.type:
            case SourceType.text | SourceType.pickle:
                return col in self._row
            case SourceType.ini:
                return self._parent.has_option(self._row, col)
            case SourceType.xml:
                if self._row.find(col) is not None:
                    return True
                return False
        return False

    def setvalue(self, col: str, value: str, /) -> None:
        """Set the value of the specified column.

        Args:
            col: The name of the column for which to set the value.
            value: The value to set for the column.

        Returns:
            Nothing.
        """
        match self.type:
            case SourceType.text | SourceType.pickle:
                self._row[col] = value
            case SourceType.ini:
                self._parent.set(self._row, col, value)
            case SourceType.xml:
                col_ref = self._row.find(col)
                if col_ref is None:
                    col_ref = xml_etree.SubElement(self._row, col)
                col_ref.text = value


class DataTable:
    """Class to create a universal abstract interface for a data table.

    The _parent meaning changes based on the source type:
        TEXT = absolute path to the data directory
        PICKLE = top level dictionary
        XML_* = the parent element

    Attributes:
        _INI_ROW_FORMAT: The format for an INI file row.
        _XML_ROW_TAG: The XML file tag to denote a row.
        _XML_SINGLE_ROW_TAG: The XML file tag to denote a row (xml_single data source).
    """
    _INI_ROW_FORMAT = f'%s{DataSource.INI_ROW_TAG}%d'
    _XML_ROW_TAG = 'ROW'
    _XML_SINGLE_ROW_TAG = 'environment'

    def __init__(self, data_type: SourceType, name: str, raw: Any, parent: Any, /):
        """
        Args:
            data_type: The data source type.
            name: The table name.
            raw: The raw value of the table as read from the data source.
            parent: This is a reference to the data source containing the table.

        Attributes:
            _name: The value of the name argument.
            _parent: The value of the parent argument.
            _table: The value of the raw argument.
            _type: The value of the data_type argument.
        """
        self._name = name
        self._parent = parent
        self._table = raw
        self._type = data_type

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def _get_row_parent(self) -> Any:
        """Get the parent of a row in this table.

        Returns:
            The parent of a row in this table.
        """
        match self.type:
            case SourceType.text | SourceType.pickle | SourceType.xml:
                return self._table
            case SourceType.ini:
                return self._parent
        raise DataError(DataError.INVALID_TYPE)

    name = property(lambda s: s._name, doc='A read-only property which returns the name of the data table.')
    raw = property(lambda s: s._table, doc='A read-only property which returns the raw contents of the data table.')
    type = property(lambda s: s._type, doc='A read-only property which returns the type of the data source.')

    def add_row(self, **values) -> DataRow:
        """Create a new row with the specified values.

        Args:
            **values: A dictionary of the value to add.

        Returns:
            Nothing.
        """
        raw_row: Any = None
        match self.type:
            case SourceType.text | SourceType.pickle | SourceType.xml:
                row_num = (int(self._table[-1]) + 1) if self._table else 1
                self._parent.add_section(self._INI_ROW_FORMAT % (self.name, row_num))
                self._table.append(row_num)
                self._parent.set(self.name, DataSource.INI_ROW_LIST_OPT, ','.join([str(r) for r in self._table]))
                raw_row = self._INI_ROW_FORMAT % (self.name, row_num)
            case SourceType.xml:
                raw_row = xml_etree.SubElement(self._table, self._XML_ROW_TAG)
        row = DataRow(self.type, raw_row, self._get_row_parent())
        for (var, val) in values.items():
            row.setvalue(var, val)
        return row

    def del_row(self, col: str, value: str, /) -> None:
        """Delete the rows with the specified column value.

        Args:
            col: The column to select for.
            value: The column value to select for.

        Returns:
            Nothing.
        """
        for row in self.get_rows(col, value):
            row.delete()

    def get_rows(self, col: Optional[str] = None, value: Optional[str] = None) -> List[DataRow]:
        """Get all the rows matching the specified selector.

        Args:
            col (optional, default=None): The column to select for if not None.
            value (optional, default=None): The column value to select for if not None.

        Returns:
            The list of rows matching the specified selector.
        """
        row_list = self._table if self.type != SourceType.ini else [self._INI_ROW_FORMAT % (self.name, r) for r in self._table]
        all_rows = [DataRow(self.type, r, self._get_row_parent()) for r in row_list]
        if (col is None) and (value is None):
            return all_rows
        if value is None:
            return [r for r in all_rows if r.has_col(str(col))]
        return [r for r in all_rows if r.get_value(str(col)) == value]

# cSpell:ignore getiterator



# flake8: noqa
# pylint: disable-all
# cSpell:ignore unindented todos numericals docstrings

# Step1: Add module level docstrings, including attributes.
# Step2: Add module level function docstrings.
# Step3: Add class docstrings.
# Step4: Add Error docstrings.
# Step5: Add class __init__ docstrings. Arguments, Attributes, Raises.
# Step6: Add class property docstrings.
# Step7: Add class method docstrings.
# Step8: Add class private method docstrings.
# Step9: Reorder Classes: private, protected, properties, methods.
# Step10: Fix enums.
# Step11: Add PEP484 type hints.
# Step12: Add PEP526 variable type annotations
# Step13: Add linting and type checking to build pipeline.
# Step14: Change pylint numericals
# Step15: Walrus operator (:=)
# Step16: Positional arguments (/, *)

# -*- coding: utf-8 -*-
"""Example Google style docstrings.

This module demonstrates documentation as specified by the `Google Python
Style Guide`_. Docstrings may extend over multiple lines. Sections are created
with a section header and a colon followed by a block of indented text.

Example:
    Examples can be given using either the ``Example`` or ``Examples``
    sections. Sections support any reStructuredText formatting, including
    literal blocks::

        $ python example_google.py

Section breaks are created by resuming unindented text. Section breaks
are also implicitly created anytime a new section starts.

Attributes:
    module_level_variable1 (int): Module level variables may be documented in
        either the ``Attributes`` section of the module docstring, or in an
        inline docstring immediately following the variable.

        Either form is acceptable, but the two should not be mixed. Choose
        one convention to document module level variables and be consistent
        with it.

Todo:
    * For module TODOs
    * You have to also use ``sphinx.ext.todo`` extension

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

module_level_variable1 = 12345


def function_with_pep484_type_annotations(param1: int, param2: str) -> bool:
    """Example function with PEP 484 type annotations.

    Args:
        param1: The first parameter.
        param2: The second parameter.

    Returns:
        The return value. True for success, False otherwise.

    """


def module_level_function(param1: int, param2:int=None, *args, **kwargs) -> bool:
    """This is an example of a module level function.

    Function parameters should be documented in the ``Args`` section. The name
    of each parameter is required. The type and description of each parameter
    is optional, but should be included if not obvious.

    If \*args or \*\*kwargs are accepted,
    they should be listed as ``*args`` and ``**kwargs``.

    The format for a parameter is::

        name: description
            The description may span multiple lines. Following
            lines should be indented. The "(type)" is optional.

            Multiple paragraphs are supported in parameter
            descriptions.

    Args:
        param1: The first parameter.
        param2: (optional, default=None): The second parameter.
            Second line of description should be indented.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        Returns True if successful, False otherwise.

        The ``Returns`` section may span multiple lines and paragraphs.
        Following lines should be indented to match the first line.

        The ``Returns`` section supports any reStructuredText formatting,
        including literal blocks::

            {
                'param1': param1,
                'param2': param2
            }

    Raises:
        AttributeError: The ``Raises`` section is a list of all exceptions
            that are relevant to the interface.
        ValueError: If `param2` is equal to `param1`.

    """
    if param1 == param2:
        raise ValueError('param1 may not be equal to param2')
    return True


def example_generator(n):
    """Generators have a ``Yields`` section instead of a ``Returns`` section.

    Args:
        n (int): The upper limit of the range to generate, from 0 to `n` - 1.

    Yields:
        int: The next number in the range of 0 to `n` - 1.

    Examples:
        Examples should be written in doctest format, and should illustrate how
        to use the function.

        >>> print([i for i in example_generator(4)])
        [0, 1, 2, 3]

    """
    for i in range(n):
        yield i


class ExampleError(Exception):
    """Exceptions are documented in the same way as classes.

    The __init__ method may be documented in either the class level
    docstring, or as a docstring on the __init__ method itself.

    Either form is acceptable, but the two should not be mixed. Choose one
    convention to document the __init__ method and be consistent with it.

    Note:
        Do not include the `self` parameter in the ``Args`` section.

    Args:
        msg (str): Human readable string describing the exception.
        code (:obj:`int`, optional): Error code.

    Attributes:
        msg (str): Human readable string describing the exception.
        code (int): Exception error code.

    """

    def __init__(self, msg, code):
        self.msg = msg
        self.code = code


class ExampleClass(object):
    """The summary line for a class docstring should fit on one line.

    If the class has public attributes, they may be documented here
    in an ``Attributes`` section and follow the same formatting as a
    function's ``Args`` section. Alternatively, attributes may be documented
    inline with the attribute's declaration (see __init__ method below).

    Properties created with the ``@property`` decorator should be documented
    in the property's getter method.

    Attributes:
        attr1 (str): Description of `attr1`.
        attr2 (:obj:`int`, optional): Description of `attr2`.

    """

    def __init__(self, param1, param2, param3):
        """Example of docstring on the __init__ method.

        The __init__ method may be documented in either the class level
        docstring, or as a docstring on the __init__ method itself.

        Either form is acceptable, but the two should not be mixed. Choose one
        convention to document the __init__ method and be consistent with it.

        Note:
            Do not include the `self` parameter in the ``Args`` section.

        Args:
            param1 (str): Description of `param1`.
            param2 (:obj:`int`, optional): Description of `param2`. Multiple
                lines are supported.
            param3 (:obj:`list` of :obj:`str`): Description of `param3`.

        """
        self.attr1 = param1
        self.attr2 = param2
        self.attr3 = param3  #: Doc comment *inline* with attribute

        #: list of str: Doc comment *before* attribute, with type specified
        self.attr4 = ['attr4']

        self.attr5 = None
        """str: Docstring *after* attribute, with type specified."""

    @property
    def readonly_property(self):
        """str: Properties should be documented in their getter method."""
        return 'readonly_property'

    @property
    def readwrite_property(self):
        """:obj:`list` of :obj:`str`: Properties with both a getter and setter
        should only be documented in their getter method.

        If the setter method contains notable behavior, it should be
        mentioned here.
        """
        return ['readwrite_property']

    @readwrite_property.setter
    def readwrite_property(self, value):
        value

    def example_method(self, param1, param2):
        """Class methods are similar to regular functions.

        Note:
            Do not include the `self` parameter in the ``Args`` section.

        Args:
            param1: The first parameter.
            param2: The second parameter.

        Returns:
            True if successful, False otherwise.

        """
        return True

    def __special__(self):
        """By default special members with docstrings are not included.

        Special members are any methods or attributes that start with and
        end with a double underscore. Any special member with a docstring
        will be included in the output, if
        ``napoleon_include_special_with_doc`` is set to True.

        This behavior can be enabled by changing the following setting in
        Sphinx's conf.py::

            napoleon_include_special_with_doc = True

        """
        pass

    def __special_without_docstring__(self):
        pass

    def _private(self):
        """By default private members are not included.

        Private members are any methods or attributes that start with an
        underscore and are *not* special. By default they are not included
        in the output.

        This behavior can be changed such that private members *are* included
        by changing the following setting in Sphinx's conf.py::

            napoleon_include_private_with_doc = True

        """
        pass

    def _private_without_docstring(self):
        pass



"""BatCave Utilities module.

This module provides utilities for making it easier to write Python programs.

The exported modules fall into several categories:

Modules that provide simplified interfaces to standard modules:
    * commander - argparse
    * platarch - platform

Modules that provide simplified interfaces to third-party modules:
    * gui - PyQt5
    * k8s - kubernetes

Modules that provide Pythonic interfaces to external systems:
    * cloudmgr - cloud resources
    * cms - code management systems
    * data - data sources
    * iispy - Internet Information Server
    * netutil - network services
    * qbpy - QuickBuild
    * servermgr - OS-independent servers
    * sysutil - system utilities
    * tcpy - TeamCity

Modules that provide utilities for accomplishing specific programming tasks:
    * automation - building automation
    * configmgr - managing configurations
    * expander - working with string expansion
    * fileutil - working with files
    * lang - Python language utilities
    * menu - creating command line menus
    * statemachine - a simple state machine
    * reporter - creating reports
    * version - reporting version information
"""

__all__ = ('__title__', '__summary__', '__uri__',
           '__version__', '__build_name__', '__build_date__',
           '__author__', '__email__',
           '__license__', '__copyright__')

__title__ = 'BatCave'
__summary__ = 'Python Programming Toolkit'
__uri__ = 'https://github.com/arisilon/batcave/'

__version__ = '47.1.0'
__build_name__ = '{var:build_name}'
__build_date__ = '{var:build_date}'

__author__ = 'Jeffery G. Smith'
__email__ = 'web@pobox.com'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2023 Jeffery G. Smith'

# cSpell:ignore iispy netutil qbpy tcpy platarch cloudmgr servermgr configmgr fileutil statemachine
